"""
Module for the disconnect dialog that prompts the user for confirmation
upon logout or exit.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Union, Callable, Optional
from gi.repository import Pango
from proton.vpn.app.gtk import Gtk
from proton.vpn import logging
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect

logger = logging.getLogger(__name__)


class ConfirmationDialog(Gtk.Dialog):
    """Confirmation dialog widget."""
    WIDTH = 150
    HEIGHT = 200

    def __init__(
        self, message: Union[Gtk.Widget, str],
        title: str,
        yes_text: Optional[str] = None, no_text: Optional[str] = None
    ):
        super().__init__()
        self.set_title(title)
        self.set_default_size(self.WIDTH, self.HEIGHT)

        yes_button = self.add_button("_Yes" if not yes_text else yes_text, Gtk.ResponseType.YES)
        no_button = self.add_button("_No" if not no_text else no_text, Gtk.ResponseType.NO)

        no_button.add_css_class("primary")
        yes_button.add_css_class("danger")

        widget: Gtk.Widget
        if isinstance(message, str):
            widget = Gtk.Label(label=message)
            widget.set_width_chars(50)
            widget.set_max_width_chars(50)
            widget.set_wrap(True)
            widget.set_wrap_mode(Pango.WrapMode.WORD)
        else:
            widget = message

        content_area: Gtk.Box = self.get_content_area()
        content_area.set_margin_top(20)
        content_area.set_margin_bottom(20)
        content_area.set_margin_start(20)
        content_area.set_margin_end(20)
        content_area.set_spacing(20)
        content_area.append(widget)


def show_confirmation_dialog(  # pylint: disable=too-many-arguments
        parent: Gtk.Widget,
        title: str,
        question: str,
        clarification: str,
        yes_text: str,
        no_text: str,
        callback_result: Callable[[ConfirmationDialog, int], None]):
    """
    Shows a confirmation dialog with a question and a clarification.
    """
    container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
    container.set_spacing(10)

    question_label = Gtk.Label(label=question)
    question_label.set_halign(Gtk.Align.START)

    clarification_label = Gtk.Label(label=clarification)
    clarification_label.set_halign(Gtk.Align.START)
    clarification_label.add_css_class("dim-label")

    container.append(question_label)
    container.append(clarification_label)

    dialog = ConfirmationDialog(
        message=container,
        title=title,
        yes_text=yes_text,
        no_text=no_text
    )
    dialog.set_default_size(400, 200)
    dialog.set_modal(True)
    dialog.set_transient_for(parent)
    safe_signal_connect(dialog, "response", callback_result)
    dialog.present()



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from types import ModuleType
from typing import Callable, Optional

from gi.repository import Gtk

from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.connection.enum import KillSwitchSetting\
    as KillSwitchSettingEnum

from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common\
    import ConflictableToggleWidget, SettingDescription, SettingName
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import (
    ReactiveSetting
)

class KillSwitchWidget(ConflictableToggleWidget, ReactiveSetting):  # noqa pylint: disable=too-many-instance-attributes,too-few-public-methods
    """Kill switch setting widget.

    Since the kill switch can have multiple modes, we need to have a proper
    widget that handles of all of these cases and is easy to test.
    """
    KILLSWITCH_LABEL = "Kill switch"
    KILLSWITCH_DESCRIPTION = "Protects your IP address by disconnecting you from the " \
        "internet if you lose your VPN connection. "\
        "<a href=\"https://protonvpn.com/support/what-is-kill-switch/\">Learn more</a> \n\n" \
        "Kill switch can only be changed when VPN is disconnected."
    KILLSWITCH_STANDARD_DESCRIPTION = "Automatically disconnect from the internet if "\
        "VPN connection is lost."
    KILLSWITCH_ADVANCED_DESCRIPTION = "Only allow internet access when connected to Proton VPN. " \
        "Advanced kill switch will remain active even when you restart your device."
    SETTING_NAME = "settings.killswitch"

    def __init__(self, controller: Controller, gtk: Optional[ModuleType] = None,
                 conflict_resolver: Optional[Callable] = None):
        super().__init__(
            controller=controller,
            title=self.KILLSWITCH_LABEL,
            description=self.KILLSWITCH_DESCRIPTION,
            setting_name=self.SETTING_NAME,
            do_set=self._do_set,
            do_revert=self._do_revert,
            disable_on_active_connection=True,
            conflict_resolver=conflict_resolver,
        )

        self.gtk = gtk or Gtk
        self._controller = controller
        self._standard_radio_button_connect_id = None
        self._advanced_radio_button_connect_id = None

        self.standard_radio_button: Optional[Gtk.CheckButton] = None
        self.advanced_radio_button: Optional[Gtk.CheckButton] = None
        self.revealer: Optional[Gtk.Revealer] = None

    def build_revealer(self):
        """Builds the revealer"""
        self.revealer = self.gtk.Revealer()
        self.attach(self.revealer, 0, 2, 2, 1)
        revealer_container = self._build_revealer_container()
        self.revealer.set_child(revealer_container)
        self.revealer.set_reveal_child(self.get_setting() > KillSwitchSettingEnum.OFF)

    @staticmethod
    def build(controller: Controller) -> "KillSwitchWidget":
        """Shortcut method to initialize widget."""
        widget = KillSwitchWidget(controller)
        widget.build_revealer()
        return widget

    def _build_revealer_container(self) -> Gtk.Box:
        # Add both containers that contain all children that are to be displayed in the revealer
        revealer_container = self.gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        revealer_container.set_spacing(10)
        revealer_container.append(self._build_standard_killswitch())
        revealer_container.append(self._build_advanced_killswitch())

        return revealer_container

    def _build_standard_killswitch(self) -> Gtk.Grid:
        main_standard_container = self.gtk.Grid()
        main_standard_container.set_column_spacing(10)

        self.standard_radio_button = Gtk.CheckButton()
        self.standard_radio_button.set_active(self.get_setting() == KillSwitchSettingEnum.ON)

        main_standard_container.attach(self.standard_radio_button, 0, 0, 1, 1)
        main_standard_container.attach(SettingName("Standard"), 1, 0, 1, 1)
        main_standard_container.attach(
            SettingDescription(self.KILLSWITCH_STANDARD_DESCRIPTION),
            1, 1, 1, 1
        )

        safe_signal_connect(
            self.standard_radio_button,
            "toggled",
            self._on_standard_radio_button_toggle
        )

        return main_standard_container

    def _build_advanced_killswitch(self) -> Gtk.Grid:
        main_advanced_container = self.gtk.Grid()
        main_advanced_container.set_column_spacing(10)

        self.advanced_radio_button = Gtk.CheckButton()
        self.advanced_radio_button.set_group(self.standard_radio_button)
        self.advanced_radio_button.set_active(self.get_setting() == KillSwitchSettingEnum.PERMANENT)

        main_advanced_container.attach(self.advanced_radio_button, 0, 0, 1, 1)
        main_advanced_container.attach(SettingName("Advanced"), 1, 0, 1, 1)
        main_advanced_container.attach(
            SettingDescription(self.KILLSWITCH_ADVANCED_DESCRIPTION),
            1, 1, 1, 1
        )

        safe_signal_connect(
            self.advanced_radio_button,
            "toggled",
            self._on_advanced_radio_button_toggle
        )

        return main_advanced_container

    def _on_standard_radio_button_toggle(self, button):
        self._handle_toggle(button, KillSwitchSettingEnum.ON)

    def _on_advanced_radio_button_toggle(self, button):
        self._handle_toggle(button, KillSwitchSettingEnum.PERMANENT)

    def _handle_toggle(self, radio_button: Gtk.CheckButton, new_value: int):
        # If revealer is hidden then we don't want to resolve the trigger from
        # programmatically setting the standard radio button.
        if not self.revealer.get_reveal_child():
            return

        if radio_button.get_active():
            self.save_setting(new_value)

    def _do_set(self, _toggle, new_value: bool):
        self.save_setting(int(new_value))
        self.revealer.set_reveal_child(new_value)

        self.standard_radio_button.set_active(True)

    def _do_revert(self, _toggle):
        self.switch.set_active(False)

    def get_killswitch_state(self) -> KillSwitchSettingEnum:
        """Returns the current kill switch state."""
        if self.switch.get_active():
            if self.standard_radio_button.get_active():
                return KillSwitchSettingEnum.ON
            if self.advanced_radio_button.get_active():
                return KillSwitchSettingEnum.PERMANENT
        return KillSwitchSettingEnum.OFF

    def set_killswitch_state(self, state: KillSwitchSettingEnum):
        """Returns the current kill switch state."""

        if state == KillSwitchSettingEnum.ON:
            self.switch.set_active(True)
            self.standard_radio_button.set_active(True)
        elif state == KillSwitchSettingEnum.PERMANENT:
            self.switch.set_active(True)
            self.advanced_radio_button.set_active(True)
        else:
            self.switch.set_active(False)

    def on_settings_changed(self, settings):
        if self.get_killswitch_state() != settings.killswitch:
            self.set_killswitch_state(settings.killswitch)



"""
Error messenger module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from dataclasses import dataclass
from typing import Optional, Callable, List

from gi.repository import GLib, Notify

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.assets.icons import ICONS_PATH
from proton.vpn.app.gtk.utils.glib import run_once
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.notification_bar import NotificationBar


@dataclass
class DialogButton:
    """A button for a dialog."""
    label: str
    response_type: Gtk.ResponseType


class Notifications:
    """It wraps all types of notifications to be shown by the app."""

    def __init__(
        self, main_window: Gtk.ApplicationWindow,
        notification_bar: NotificationBar,
    ):
        self._main_window = main_window
        self.notification_bar = notification_bar
        self.error_dialog: Optional[Gtk.MessageDialog] = None
        self._on_dialog_closed = None
        Notify.init("Proton VPN")

    def show_error_dialog(
            self, message: str, title: str, hint: Optional[str] = None,
            message_type: Gtk.MessageType = Gtk.MessageType.ERROR,
            buttons: Optional[List[DialogButton]] = None,
            on_dialog_closed: Optional[Callable] = None
    ):  # pylint: disable=too-many-arguments
        """Show an error dialog to the user."""
        run_once(
            self._generate_and_show_dialog, title, message, hint, message_type,
            buttons, on_dialog_closed
        )

    def _generate_and_show_dialog(
            self, title: str, message: str, hint: Optional[str],
            message_type: Optional[Gtk.MessageType],
            buttons: Optional[List[DialogButton]],
            on_dialog_closed: Optional[Callable]
    ):  # pylint: disable=too-many-arguments
        """Generates and displays a pop-up dialog to the user, blocking
        the rest of the UI."""
        if self.error_dialog:
            self.error_dialog.destroy()
            self.error_dialog = None

        self.error_dialog = Gtk.MessageDialog(
            transient_for=self._main_window,
            message_type=message_type,
            buttons=Gtk.ButtonsType.NONE,
            text=title,
        )
        self.error_dialog.set_destroy_with_parent(True)

        buttons = buttons or [DialogButton("OK", Gtk.ResponseType.OK)]
        for button in buttons:
            self.error_dialog.add_button(button.label, button.response_type)

        secondary_text = message
        if hint:
            secondary_text += f"\n\n<span size=\"smaller\" weight=\"light\">{hint}</span>"

        self.error_dialog.set_modal(True)
        self.error_dialog.set_markup(secondary_text)

        self._on_dialog_closed = on_dialog_closed
        safe_signal_connect(self.error_dialog, "response", self._on_dialog_response)
        self.error_dialog.present()

    def _on_dialog_response(self, dialog, response_id):
        dialog.destroy()
        self.error_dialog = None
        if self._on_dialog_closed:
            run_once(self._on_dialog_closed, response_id)
            self._on_dialog_closed = None

    def show_error_message(self, message: str):
        """Shows the error message in the notification bar."""
        GLib.idle_add(
            self.notification_bar.show_error_message,
            message
        )

    def show_info_message(self, message: str):
        """Shows the info message in the notification bar."""
        GLib.idle_add(
            self.notification_bar.show_info_message,
            message
        )

    def show_success_message(self, message: str):
        """Shows a success message in the notification bar."""
        GLib.idle_add(
            self.notification_bar.show_success_message,
            message
        )

    def hide_message(self):
        """Hides the revealed content."""
        GLib.idle_add(
            self.notification_bar.clear
        )

    def show_gnome_notification(self, title: str, description: str):
        """Shows a gnome desktop notification."""
        Notify.Notification.new(
            title, description,
            str(ICONS_PATH / "proton-vpn-sign.svg")
        ).show()



"""
This module contains common objects that are used by different settings types.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from contextlib import contextmanager
from typing import List, Tuple, Callable, Any, Optional, TYPE_CHECKING, cast
from gi.repository import Gtk, Gio

from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.confirmation_dialog import \
    ConfirmationDialog,\
    show_confirmation_dialog
from proton.vpn.app.gtk.controller import Controller
from proton.vpn import logging

if TYPE_CHECKING:
    from proton.vpn.app.gtk.conflicts import Conflict

logger = logging.getLogger(__name__)

RECONNECT_MESSAGE = "Please establish a new VPN connection for "\
        "changes to take effect."


class CategoryHeader(Gtk.Label):
    """Header that is used to seperate between setting types, such as
    Features, Connection, etc."""
    def __init__(self, label: str):
        super().__init__(label=label)
        self.set_halign(Gtk.Align.START)
        self.add_css_class("heading")


class ReactiveSetting:  # pylint: disable=too-few-public-methods
    """Base class for reactive settings that need to be updated when settings change."""
    def on_settings_changed(self, _settings):
        """Method that is called when settings are changed.
        This is used to update the widget when settings change."""
        raise NotImplementedError("This method should be implemented in subclasses.")


class ReactiveSettingContainer:  # pylint: disable=too-few-public-methods
    """Base class for Gtk containers that hold reactive settings."""
    def on_settings_changed(self, settings):
        """Method that is called when settings are changed.
        This is used to update the widget when settings change."""
        child = self.get_first_child()
        while child:
            if isinstance(child, ReactiveSetting):
                child.on_settings_changed(settings)
            child = child.get_next_sibling()  # pylint: disable=no-member

    def get_first_child(self):
        """Returns the first child of the container.
        It can be then used to iterate through the children of the container with
        ```
        child = self.get_first_child()
        next_sibling = child.get_next_sibling()
        ```
        """
        return NotImplementedError("This method should be implemented.")


class BaseCategoryContainer(Gtk.Box):
    """Base container class that is used to group common settings.

    Mostly a helper class to remove the necessity of writing boilerplate
    styling code in each category container.
    """
    def __init__(self, category_name: str):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        self.add_css_class("setting-category")
        self.set_halign(Gtk.Align.FILL)
        self.set_spacing(15)

        self.append(CategoryHeader(category_name))


class BetaTag(Gtk.Label):
    """A label styled with a purple border to indicate a beta feature."""
    LABEL = "BETA"

    def __init__(self):
        super().__init__(label=self.LABEL)
        self.add_css_class("beta-tag")
        self.set_valign(Gtk.Align.CENTER)


class UpgradePlusTag(Gtk.Button):
    """ Using normal button instead of LinkButton mainly
    because of styling. LinkButtons usually have very ugly UI,
    and all they do is emit `::activate-link` which
    just calls `Gtk.show_uri_on_window`.

    Source: https://lazka.github.io/pgi-docs/Gtk-3.0/classes/LinkButton.html
    """
    LABEL = "VPN Plus"
    URL = "https://protonvpn.com/pricing"

    def __init__(self):
        super().__init__(label=self.LABEL)
        self.add_css_class("upgrade-tag")
        self.add_css_class("heading")
        safe_signal_connect(self, "clicked", self._on_button_clicked)

    def _on_button_clicked(self, _):
        Gio.AppInfo.launch_default_for_uri(self.URL, None)


class SettingName(Gtk.Label):
    """Label used to identify a setting."""
    def __init__(self, label: str, bold: bool = False):
        if bold:
            label = f"<b>{label}</b>"

        super().__init__(label=label)
        self.set_halign(Gtk.Align.START)
        self.set_hexpand(True)
        self.set_use_markup(True)

    @property
    def enabled(self) -> bool:
        """Returns if the label is enabled."""
        return self.get_property("sensitive")

    @enabled.setter
    def enabled(self, new_value: bool):
        """Sets if the label should be enabled."""
        self.set_property("sensitive", new_value)


class SettingDescription(Gtk.Label):
    """Label used to desribe a setting."""
    def __init__(self, label: str):
        super().__init__(label=label)
        self.add_css_class("dim-label")
        self.set_wrap(True)
        self.set_max_width_chars(1)
        self.set_property("xalign", 0)
        self.set_hexpand(True)
        self.set_use_markup(True)


def is_upgrade_required(requires_subscription_to_be_active: bool, user_tier: int) -> bool:
    """Returns if an upgrade is required for a certain setting."""
    return requires_subscription_to_be_active and user_tier < 1


class CustomButton(Gtk.Grid):
    """Custom button setting."""
    def __init__(  # pylint: disable=too-many-arguments
        self,
        title: str,
        description: str,
        button_label: str,
        on_click_callback: Callable,
        requires_subscription_to_be_active: bool = False,
        bold_title: bool = False
    ):
        super().__init__()
        self._apply_grid_styles()
        self._requires_subscription_to_be_active = requires_subscription_to_be_active
        self.label = SettingName(title, bold=bold_title)
        self.description = SettingDescription(description)
        self.button = self._build_button(button_label, on_click_callback)
        self._build_ui()

    def _build_button(self, button_label: str, on_click_callback: Callable) -> Gtk.Button:
        button = Gtk.Button()
        button.set_label(button_label)
        safe_signal_connect(button, "clicked", on_click_callback)
        return button

    def _build_ui(self):
        """Builds the UI depending if an upgrade is required or not."""
        if self._requires_subscription_to_be_active:
            self.button = UpgradePlusTag()

        self.attach(self.label, 0, 0, 1, 1)

        # Style interactive_object so it's always aligned
        self.button.set_hexpand(True)
        self.button.set_halign(Gtk.Align.END)

        self.attach(self.button, 1, 0, 1, 1)

        if self.description:
            self.attach(self.description, 0, 1, 2, 1)

    def _apply_grid_styles(self):
        self.add_css_class("setting-item")
        self.set_halign(Gtk.Align.FILL)
        self.set_row_spacing(10)
        self.set_column_spacing(100)


class ToggleWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
    """Default toggle widget."""
    def __init__(  # pylint: disable=too-many-arguments
        self,
        controller: Controller,
        title: str,
        description: str,
        setting_name: str,
        requires_subscription_to_be_active: bool = False,
        callback: Optional[Callable] = None,
        disable_on_active_connection: bool = False,
        enabled: Optional[bool] = None,
        display_tooltip_only_on_active_connection: bool = False
    ):
        super().__init__()
        self._apply_grid_styles()
        self._controller = controller
        self._setting_name = setting_name
        self._callback = callback
        self._requires_subscription_to_be_active = requires_subscription_to_be_active
        self._disable_on_active_connection = disable_on_active_connection
        self._enabled = enabled
        self._display_tooltip_only_on_active_connection = display_tooltip_only_on_active_connection

        self.label = SettingName(title)
        self.description = SettingDescription(description)
        self.switch = self._build_switch()

        self._build_ui()

    @property
    def active(self) -> bool:
        """Returns if the widget is active or not."""
        return self.get_property("sensitive")

    @active.setter
    def active(self, new_value: bool):
        """Set if the widget should be active or not."""
        self.set_property("sensitive", new_value)

    def get_setting(self) -> bool:
        """Shortcut property that returns the current setting"""
        return cast(bool, self._controller.get_setting_attr(self._setting_name))

    def save_setting(self, new_value: bool):
        """Shortcut property that sets the new setting and stores to disk."""
        self._controller.save_setting_attr(self._setting_name, new_value)

    @property
    def overridden_by_upgrade_tag(self) -> bool:
        """Returns if the the upgrade tag has overridden original interactive
        object."""
        return isinstance(self.switch, UpgradePlusTag)

    def set_tooltip(self, tooltip_text: str):
        """Set a tooltip to this row."""
        if (
            self._display_tooltip_only_on_active_connection
            and not self._controller.connection_disconnected
        ):
            self.set_has_tooltip(True)
            self.set_tooltip_text(tooltip_text)

    def _apply_grid_styles(self):
        self.add_css_class("setting-item")
        self.set_halign(Gtk.Align.FILL)
        self.set_row_spacing(10)
        self.set_column_spacing(100)

    def _build_switch(self) -> Gtk.Switch:
        switch = Gtk.Switch()
        if self._enabled is None:
            self._enabled = self.get_setting()

        switch.set_active(self._enabled)

        safe_signal_connect(switch, "notify::active", self._on_switch_state)

        return switch

    def _build_ui(self):
        """Builds the UI depending if an upgrade is required or not."""
        if self._is_upgrade_required:
            self.switch = UpgradePlusTag()

        self.attach(self.label, 0, 0, 1, 1)

        # Style interactive_object so it's always aligned
        self.switch.set_hexpand(True)
        self.switch.set_halign(Gtk.Align.END)

        self.attach(self.switch, 1, 0, 1, 1)

        if self.description:
            self.attach(self.description, 0, 1, 2, 1)

        if not self._controller.connection_disconnected and self._disable_on_active_connection:
            self.active = False

    def _on_switch_state(self, switch, _gparam):
        new_value = switch.get_active()
        if self._callback:
            self._callback(self, new_value, self)
        else:
            self.save_setting(new_value)

    @property
    def _is_upgrade_required(self) -> bool:
        """Returns if an upgrade is required for a given setting."""
        return is_upgrade_required(
            self._requires_subscription_to_be_active,
            self._controller.user_tier
        )

    def off(self):
        """Shortcut to toggle the widget to disabled."""
        self.switch.set_active(False)

    def set_state(self, state: bool):
        """Sets the switch state."""
        self.switch.set_state(state)


class ConflictableToggleWidget(ToggleWidget):  # pylint: disable=too-many-instance-attributes
    """
    Toggle widget that can handle conflicts when toggling the switch.
    It will show a confirmation dialog if the toggle is in conflict with
    other settings.
    """
    def __init__(  # pylint: disable=too-many-arguments
        self,
        controller: Controller,
        title: str,
        description: str,
        setting_name: str,
        do_set: Callable[[ToggleWidget, bool], None],
        do_revert: Callable[[ToggleWidget], None],
        requires_subscription: bool = False,
        disable_on_active_connection: bool = False,
        enabled: Optional[bool] = None,
        display_tooltip_only_on_active_connection: bool = False,
        conflict_resolver: Optional[Callable[[str, Any], Optional["Conflict"]]] = None,
    ):
        super().__init__(
            controller=controller, title=title,
            description=description, setting_name=setting_name,
            requires_subscription_to_be_active=requires_subscription,
            callback=self._on_switch_button_toggle,
            disable_on_active_connection=disable_on_active_connection,
            enabled=enabled,
            display_tooltip_only_on_active_connection=display_tooltip_only_on_active_connection
        )
        self._new_value = None
        self.do_set = do_set
        self.do_revert = do_revert

        # Allow to pass a custom conflict resolver
        if conflict_resolver is None:
            conflict_resolver = controller.setting_attr_has_conflict
        self._conflict_resolver = conflict_resolver

    def _on_switch_button_toggle(self, _, new_value: bool, __):

        if conflict := self._conflict_resolver(self._setting_name, new_value):

            self._new_value = new_value
            show_confirmation_dialog(
                self.get_root(),
                title="",
                question=conflict.label,
                clarification=conflict.description,
                yes_text="_Yes",
                no_text="_Cancel",
                callback_result=self._confirm_change
            )
        else:
            self.do_set(self, new_value)

    def _confirm_change(self, dialog: ConfirmationDialog, response: int):
        if response == Gtk.ResponseType.YES:
            # do_set can be any callable that takes
            # (ToggleWidget, int) as arguments so although it looks
            # like we are calling a method on self, we are actually
            # calling the method that was passed to the constructor.
            # This allows for instantiating this class with different
            # do_set and do_revert methods without having to
            # subclass it.
            #
            # We are passing `self` as the first argument
            # because the do_set method expects a ToggleWidget.
            self.do_set(self, self._new_value)
        else:
            # Similarly to above, we are calling the
            # do_revert method that was passed to the constructor,
            # it's not a method of this class.
            # This is why we are passing `self` as the first argument.
            self.do_revert(self)

        self._new_value = None
        # We cant just close the dialog, instead we destroy it
        # directly.
        dialog.destroy()


class ComboboxWidget(Gtk.Grid):  # pylint: disable=too-many-instance-attributes
    """Default combobox text widget."""
    def __init__(  # pylint: disable=too-many-arguments
        self,
        controller: Controller,
        title: str,
        setting_name: str,
        combobox_options: List[Tuple[int, str]],
        description: Optional[str] = None,
        requires_subscription_to_be_active: bool = False,
        callback: Optional[Callable] = None,
        disable_on_active_connection: bool = False
    ):
        super().__init__()
        self._apply_grid_styles()
        self._controller = controller
        self._setting_name = setting_name
        self._combobox_options = combobox_options
        self._callback = callback
        self._requires_subscription_to_be_active = requires_subscription_to_be_active
        self.label = SettingName(title)
        self.description = None if not description else SettingDescription(description)
        self._disable_on_active_connection = disable_on_active_connection
        self._changed_callback_id = None
        self.combobox = self._build_combobox()
        self._build_ui()

    @property
    def active(self) -> bool:
        """Returns if the widget is active or not."""
        return self.get_property("sensitive")

    @active.setter
    def active(self, new_value: bool):
        """Set if the widget should be active or not."""
        self.set_property("sensitive", new_value)

    def get_setting(self) -> str:
        """Shortcut property that returns the current setting"""
        return str(self._controller.get_setting_attr(self._setting_name))

    def save_setting(self, new_value: int):
        """Shortcut property that sets the new setting and stores to disk."""
        self._controller.save_setting_attr(self._setting_name, new_value)

    @property
    def overridden_by_upgrade_tag(self) -> bool:
        """Returns if the the upgrade tag has overridden original interactive
        object."""
        return isinstance(self.combobox, UpgradePlusTag)

    def set_tooltip(self, tooltip_text: str):
        """Set a tooltip to this row."""
        self.set_has_tooltip(True)
        self.set_tooltip_text(tooltip_text)

    def _apply_grid_styles(self):
        self.add_css_class("setting-item")
        self.set_halign(Gtk.Align.FILL)
        self.set_row_spacing(10)
        self.set_column_spacing(100)

    def _build_combobox(self) -> Gtk.ComboBoxText:
        combobox = Gtk.ComboBoxText()
        for value, display in self._combobox_options:
            combobox.append(str(value), display)

        combobox.set_entry_text_column(1)
        combobox.set_active_id(self.get_setting())

        if self._callback:
            self._changed_callback_id = \
                safe_signal_connect(combobox, "changed", self._callback)
        else:
            self._changed_callback_id = \
                safe_signal_connect(combobox, "changed", self._on_combobox_change)

        if not self._controller.connection_disconnected and self._disable_on_active_connection:
            self.active = False

        return combobox

    def _build_ui(self):
        """Builds the UI depending if an upgrade is required or not."""
        if self._is_upgrade_required:
            self.combobox = UpgradePlusTag()

        self.attach(self.label, 0, 0, 1, 1)

        # Style interactive_object so it's always aligned
        self.combobox.set_hexpand(True)
        self.combobox.set_halign(Gtk.Align.END)

        self.attach(self.combobox, 1, 0, 1, 1)

        if self.description:
            self.attach(self.description, 0, 1, 2, 1)

    @property
    def _is_upgrade_required(self) -> bool:
        """Returns if an upgrade is required for a given setting."""
        return is_upgrade_required(
            self._requires_subscription_to_be_active,
            self._controller.user_tier
        )

    @contextmanager
    def pause_callback(self):
        """Context manager that temporarily blocks the combobox 'changed' signal."""
        self.combobox.handler_block(self._changed_callback_id)
        try:
            yield
        finally:
            self.combobox.handler_unblock(self._changed_callback_id)

    def _on_combobox_change(self, combobox: Gtk.ComboBox):
        model = combobox.get_model()
        treeiter = combobox.get_active_iter()
        value = model[treeiter][1]
        self.save_setting(value)

    def off(self):
        """Shortcut to set the combobox to disabled."""
        self.combobox.set_active_id(str(0))


class ConflictableComboboxWidget(ComboboxWidget):
    """
    Combobox widget that can handle conflicts when changing the combobox value.
    It will show a confirmation dialog if the combobox change is in conflict with
    other settings.
    """
    def __init__(  # pylint: disable=too-many-arguments
        self,
        controller: Controller,
        title: str,
        setting_name: str,
        combobox_options: List[Tuple[int, str]],
        do_set: Callable[[ComboboxWidget, int], None],
        do_revert: Callable[[ComboboxWidget], None],
        description: Optional[str] = None,
        requires_subscription: bool = False,
        disable_on_active_connection: bool = False
    ):
        super().__init__(
            controller=controller, title=title, setting_name=setting_name,
            combobox_options=combobox_options, description=description,
            requires_subscription_to_be_active=requires_subscription,
            callback=self._on_conflictable_combobox_change,
            disable_on_active_connection=disable_on_active_connection
        )
        self._new_value = None
        self._do_set = do_set
        self._do_revert = do_revert

    def _on_conflictable_combobox_change(
            self, combobox_text_widget
    ):
        new_value = combobox_text_widget.get_active_text()

        if conflict := self._controller.setting_attr_has_conflict(
                self._setting_name, new_value):

            self._new_value = new_value
            show_confirmation_dialog(
                self.get_root(),
                title="",
                question=conflict.label,
                clarification=conflict.description,
                yes_text="_Yes",
                no_text="_Cancel",
                callback_result=self._confirm_change
            )
        else:
            self._do_set(self, new_value)

    def _confirm_change(self, dialog: ConfirmationDialog, response: int):
        if response == Gtk.ResponseType.YES:
            # do_set can be any callable that takes
            # (ComboboxWidget, int) as arguments so although it looks
            # like we are calling a method on self, we are actually
            # calling the method that was passed to the constructor.
            # This allows for instantiating this class with different
            # do_set and do_revert methods without having to
            # subclass it.
            #
            # We are passing `self` as the first argument
            # because the do_set method expects a ComboboxWidget.
            self._do_set(self, self._new_value)
        else:
            # Similarly to above, we are calling the
            # do_revert method that was passed to the constructor,
            # it's not a method of this class.
            # This is why we are passing `self` as the first argument.
            self._do_revert(self)

        self._new_value = None
        # We cant just close the dialog, instead we destroy it
        # directly.
        dialog.destroy()


class EntryWidget(Gtk.Grid):
    """Default entry widget."""
    def __init__(  # pylint: disable=too-many-arguments
        self,
        controller: Controller,
        title: str,
        setting_name: str,
        description: str,
        callback: Optional[Callable] = None,
        requires_subscription_to_be_active: bool = False,
    ):
        super().__init__()
        self._apply_grid_styles()
        self._controller = controller
        self._setting_name = setting_name
        self._callback = callback
        self._requires_subscription_to_be_active = requires_subscription_to_be_active
        self.label = SettingName(title)
        self.description = SettingDescription(description)
        self.entry = self._build_entry()
        self._build_ui()

    @property
    def active(self) -> bool:
        """Returns if the widget is active or not."""
        return self.get_property("sensitive")

    @active.setter
    def active(self, new_value: bool):
        """Set if the widget should be active or not."""
        self.set_property("sensitive", new_value)

    def get_setting(self) -> object:
        """Shortcut property that returns the current setting"""
        return self._controller.get_setting_attr(self._setting_name)

    def save_setting(self, new_value: bool):
        """Shortcut property that sets the new setting and stores to disk."""
        self._controller.save_setting_attr(self._setting_name, new_value)

    @property
    def overridden_by_upgrade_tag(self) -> bool:
        """Returns if the the upgrade tag has overridden original interactive
        object."""
        return isinstance(self.entry, UpgradePlusTag)

    def set_tooltip(self, tooltip_text: str):
        """Set a tooltip to this row."""
        self.set_has_tooltip(True)
        self.set_tooltip_text(tooltip_text)

    def _apply_grid_styles(self):
        self.add_css_class("setting-item")
        self.set_halign(Gtk.Align.FILL)
        self.set_row_spacing(10)
        self.set_column_spacing(100)

    def _build_entry(self) -> Gtk.Entry:
        entry = Gtk.Entry()
        value = self.get_setting()
        if value is None:
            value = "Off"

        entry.set_text(str(value))
        if self._callback:
            safe_signal_connect(entry, "changed", self._forward_to_callback)
        else:
            safe_signal_connect(entry, "changed", self._on_changed_event)
        return entry

    def _forward_to_callback(self, *args):
        self._callback(self.entry, self, *args)

    def change_value(self, new_value: str):
        """Change the value of the entry widget. Method added for testing purposes."""
        self.entry.set_text(new_value)
        if self._callback:
            controllers = self.entry.observe_controllers()
            for controller in controllers:
                if isinstance(controller, Gtk.EventControllerFocus):
                    controller.emit("leave")

    def _build_ui(self):
        """Builds the UI depending if an upgrade is required or not."""
        if self._is_upgrade_required:
            self.entry = UpgradePlusTag()

        self.attach(self.label, 0, 0, 1, 1)

        # Style interactive_object so it's always aligned
        self.entry.set_hexpand(True)
        self.entry.set_halign(Gtk.Align.END)

        self.attach(self.entry, 1, 0, 1, 1)

        if self.description:
            self.attach(self.description, 0, 1, 2, 1)

    def _on_changed_event(self, gtk_widget: Gtk.Entry, *_):
        self.save_setting(gtk_widget.get_text())

    @property
    def _is_upgrade_required(self) -> bool:
        """Returns if an upgrade is required for a given setting."""
        return is_upgrade_required(
            self._requires_subscription_to_be_active,
            self._controller.user_tier
        )



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Optional
import html
import re
import shutil

from gi.repository import Gio
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.app.data_structures \
    import AppData
from proton.vpn.app.gtk.util import APPLICATION_ID
from proton.vpn import logging

logger = logging.getLogger(__name__)

PROTON_VPN_APP_ID_DOT_DESKTOP = f"{APPLICATION_ID}.desktop"


# This regex is used to extract the executable from a flatpak app's `.desktop` file.
# It matches the pattern of a flatpak app's executable.
# The regex captures the app start command.
# Example:
#   `/usr/bin/flatpak run --branch=stable --arch=x86_64 --command=mock-app @@ %u @@`
# The captured group will be:
#   `/usr/bin/flatpak run --branch=stable --arch=x86_64 --command=mock-app`
# Some flatpak apps might not have the `@@ ... @@`.
FLATPAK_PATTERN = re.compile(r"^(.+?)(?:\s*@@|$)")


class DesktopFileParsingError(Exception):
    """Error while parsing desktop file."""


def _check_is_flatpak(app: Gio.AppInfo) -> bool:
    """Check if executable is a flatpak. """
    command_line = app.get_commandline()
    if command_line is None:
        return False
    return command_line.startswith("flatpak") or command_line.startswith("/usr/bin/flatpak")


def _get_flatpak_executable(app: Gio.AppInfo) -> str:
    """
    Returns the flatpak command line string trimming file forwarding
    specified with @@ ... @@. See FLATPAK_PATTERN regex.
    More info:
    https://unix.stackexchange.com/questions/797031/what-does-u-and-mean-in-the-a-desktop-entry
    https://docs.flatpak.org/en/latest/flatpak-command-reference.html
    """
    command_line = app.get_commandline()
    result = FLATPAK_PATTERN.search(command_line)

    if not result:
        raise DesktopFileParsingError(
            "Could not parse flatpack executable string from: {command_line}"
        )

    return result.group(1).rstrip()


def _check_is_snap(app: Gio.AppInfo) -> bool:
    """Checks if the command line string runs a snap app."""
    command_line = app.get_commandline()
    if command_line is None:
        return False
    return command_line.startswith("/snap/bin/")


def _get_snap_executable(app: Gio.AppInfo) -> str:
    """
    Transforms "/snap/bin/<app-name> ..." to "/snap/<app-name>/".
    Returns the transformed string or None if the exe string can't be parsed.
    """
    command_line = app.get_commandline()
    re_result = re.search(r"/snap/bin/([\w\-.]*).", command_line)

    if not re_result:
        raise DesktopFileParsingError(f"Could not parse snap app ID from: {command_line}")

    # From `/snap/bin/<app-name>` will return `<app-name>`
    app_name = re_result.group(1)

    return f"/snap/{app_name}/"


def _check_is_command(command_line: str) -> bool:
    """
    Checks if command line contains a command (i.e. argv[0] is not an absolute path).
    :returns: True if it's a command, and False otherwise.
    """
    command, *_ = command_line.split()
    return "/" not in command and bool(shutil.which(command))


def _get_native_app_executable(app: Gio.AppInfo):
    """Gets the full exe path for a command (+args)."""
    executable = app.get_executable()

    if _check_is_command(executable):
        executable: Optional[str] = shutil.which(executable)

    if not executable:
        raise DesktopFileParsingError(
            "Could not get path from command: {command}"
        )

    return executable


def get_app_icon(app: Gio.AppInfo) -> Optional[str]:
    """Returns either the name of a themed icon or the full path to the icon."""
    icon = None

    received_icon: Optional[Gio.Icon] = app.get_icon()
    if isinstance(received_icon, Gio.ThemedIcon):
        icon = received_icon.get_names()[0]
    elif isinstance(received_icon, Gio.FileIcon):
        icon = received_icon.get_file().get_path()

    return icon


def get_app_executable(app: Gio.AppInfo) -> str:
    """
    Returns the executable to split tunnel, transformed if necessary.
    """

    if _check_is_flatpak(app):
        executable = _get_flatpak_executable(app)
    elif _check_is_snap(app):
        executable = _get_snap_executable(app)
    else:
        executable = _get_native_app_executable(app)

    return executable


def get_all_installed_apps() -> list[AppData]:
    """Gets a list of installed applications on the system."""
    app_list = []

    for app in Gio.AppInfo.get_all():
        if not app.should_show():
            continue

        # Let's not split tunnel the vpn client itself,
        # otherwise it's not possible to connect to the VPN.
        if app.get_id() == PROTON_VPN_APP_ID_DOT_DESKTOP:
            continue

        try:
            executable = get_app_executable(app)
        except DesktopFileParsingError:
            logger.warning(
                "Could not get app executable for %s", app.get_filename(),
                exc_info=True
            )
            continue

        icon = get_app_icon(app)

        app_list.append(AppData(
            name=html.escape(app.get_display_name()),
            executable=executable,
            icon_name=icon
        ))

    return sorted(app_list, key=lambda app: app.name)



"""
This module defines the row content displayed in the server list widget.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import annotations
from typing import List, Optional, Set, Tuple
from gi.repository import GObject

from proton.vpn import logging

from proton.vpn.session.servers import ServerFeatureEnum

from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.row_view_model import RowViewModel

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.utils.accessibility import add_accessibility, remove_accessibility
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.vpn.serverlist.icons import (
    LocationIcon, P2PIcon, SecureCoreIcon, SmartRoutingIcon, TORIcon, UnderMaintenanceIcon
)

from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.server_load import ServerLoad
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.hover_stack import HoverStack

logger = logging.getLogger(__name__)

UPGRADE_URL = "https://account.protonvpn.com/"


class RowContent(Gtk.Box):  # pylint: disable=too-many-instance-attributes
    """Row content in the server list."""
    # pylint: disable=too-many-arguments,too-many-statements

    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL)
        self.add_css_class("row-content")
        self._connected_signals: List[Tuple[int, Gtk.Widget]] = []

        # Properties
        self._row_data: Optional[RowViewModel] = None
        self._expanded: Optional[bool] = None
        self._icon_leave: Optional[Gtk.Widget] = None
        self._icon_hover: Optional[Gtk.Widget] = None

        # UI widgets
        self._feature_icons: List[Gtk.Image] = []
        self.set_spacing(10)

        # _leave_box contains widget shown when the row doesn't have focus nor it's hovered
        self._leave_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        self._leave_box.add_css_class("row-leave-box")

        self._label = Gtk.Label()
        self._label.set_halign(Gtk.Align.START)
        self._label.set_hexpand(True)

        self._feature_icons_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self._feature_icons_box.set_spacing(10)
        self._feature_icons_box.set_halign(Gtk.Align.END)

        self._server_load = ServerLoad(0)
        self._server_load.set_visible(False)

        self._leave_box.append(self._label)
        self._leave_box.append(self._feature_icons_box)
        self._leave_box.append(self._server_load)

        # --- Hover child: full-width flat button ---
        # Mirrors _label from the leave child: shows the row name on the left.
        self._hover_name_label = Gtk.Label()
        self._hover_name_label.set_halign(Gtk.Align.START)
        self._hover_name_label.set_hexpand(True)

        # Shows the action text ("Connect" or "Upgrade") on the right.
        self._hover_action_label = Gtk.Label()
        self._hover_action_label.set_halign(Gtk.Align.END)
        self._hover_action_label.add_css_class("connect-action-label")

        self._hover_button_content = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, spacing=10
        )
        self._hover_button_content.add_css_class("row-hover-box")
        self._hover_button_content.append(self._hover_name_label)
        self._hover_button_content.append(self._hover_action_label)

        self.action_button = Gtk.Button()
        self.action_button.add_css_class("flat")
        self.action_button.add_css_class("connect-button")
        self.action_button.set_hexpand(True)
        self.action_button.set_child(self._hover_button_content)

        # Hidden label for accessibility
        self._accessibility_label = Gtk.Label()
        self._accessibility_label.set_visible(False)

        self._hover_stack = HoverStack()
        self._hover_stack.add_css_class("row-hover-stack")
        self._hover_stack.set_hexpand(True)
        self.append(self._hover_stack)
        self.append(self._accessibility_label)

        # --- Under maintenance icon ---
        self.under_maintenance_icon = UnderMaintenanceIcon()
        self.under_maintenance_icon.add_css_class("under-maintenance-icon")
        self.under_maintenance_icon.set_halign(Gtk.Align.END)
        self.under_maintenance_icon.set_hexpand(True)
        self.under_maintenance_icon.set_visible(False)
        self.append(self.under_maintenance_icon)

        # --- Toggle button (outside stack, always at the end) ---
        self._collapsed_img = Gtk.Image.new_from_icon_name("pan-down-symbolic")
        self._expanded_img = Gtk.Image.new_from_icon_name("pan-up-symbolic")
        self.toggle_button = Gtk.Button()
        self.toggle_button.add_css_class("toggle-button")
        self.toggle_button.add_css_class("flat")
        self._set_toggle_button_visible(False)
        self.append(self.toggle_button)

        safe_signal_connect(self, "unrealize", self._on_unrealize)

    def display(self, row_data: RowViewModel):
        """Displays the row content according to the specified parameters."""
        self.reset()
        self._row_data = row_data

        # Create two independent icon instances: one for leave, one for hover.
        if row_data.icon_factory is not None:
            self._icon_leave = row_data.icon_factory()
            self._icon_hover = row_data.icon_factory()
        else:
            self._icon_leave = Gtk.Image()
            self._icon_leave.set_size_request(LocationIcon.SIZE, LocationIcon.SIZE)
            self._icon_hover = Gtk.Image()
            self._icon_hover.set_size_request(LocationIcon.SIZE, LocationIcon.SIZE)

        self._leave_box.prepend(self._icon_leave)
        self._hover_button_content.prepend(self._icon_hover)
        self._hover_name_label.set_text(row_data.name)

        if row_data.load is not None:
            self._server_load.set_visible(True)
            self._server_load.set_load(row_data.load)
        else:
            self._server_load.set_visible(False)

        self._label.set_text(row_data.name)

        feature_icons = self._build_feature_icons(row_data)
        self._feature_icons = feature_icons
        for feature_icon in feature_icons:
            self._feature_icons_box.prepend(feature_icon)
        self._hover_stack.set_leave_child(self._leave_box)

        if self._row_data.under_maintenance:
            self._show_under_maintenance_icon()
        else:
            self._show_hover_stack()
            self._configure_connect_button(row_data.connect_button_tooltip)

        if row_data.toggable:
            self._set_toggle_button_visible(True)
            signal_id = safe_signal_connect(
                self.toggle_button,
                "clicked",
                self._on_toggle_button_clicked
            )
            self._connected_signals.append((signal_id, self.toggle_button))
            self.expanded = False
        else:
            self._set_toggle_button_visible(False)

    def _show_under_maintenance_icon(self):
        self.under_maintenance_icon.set_visible(True)
        help_text = f"{self.label} is under maintenance"
        self.under_maintenance_icon.set_help_text(help_text)
        self.add_css_class("dimmed")

    def _show_hover_stack(self):
        self.under_maintenance_icon.set_visible(False)
        self._hover_stack.set_visible(True)
        if self._row_data.upgrade_required:
            self._hover_action_label.set_text("Upgrade")
            signal_id = safe_signal_connect(
                self.action_button,
                "clicked",
                self._on_upgrade_button_clicked
            )
            self.add_css_class("dimmed")
        else:
            self._hover_action_label.set_text("Connect")
            signal_id = safe_signal_connect(
                self.action_button,
                "clicked",
                self._on_connect_button_clicked
            )
        self._connected_signals.append((signal_id, self.action_button))
        self._hover_stack.set_hover_child(self.action_button)

    def _configure_connect_button(self, tooltip: str):
        """Configures the connect button: accessibility, signals, and event handlers."""
        self.action_button.set_tooltip_text(tooltip)
        self._accessibility_label.set_text(tooltip)
        add_accessibility(
            self.action_button,
            Gtk.AccessibleRelation.LABELLED_BY,
            self._accessibility_label
        )
        if self._feature_icons:
            add_accessibility(
                self.action_button,
                Gtk.AccessibleRelation.DESCRIBED_BY,
                self._feature_icons
            )

    def _build_feature_icons(self, row_data: RowViewModel) -> List[Gtk.Image]:
        feature_icons: List[Gtk.Image] = []
        if ServerFeatureEnum.SECURE_CORE in row_data.features:
            if row_data.secure_core_countries:
                entry, exit_ = row_data.secure_core_countries
                feature_icons.append(SecureCoreIcon(entry, exit_))
            else:
                feature_icons.append(SecureCoreIcon())
        if row_data.smart_routing:
            feature_icons.append(SmartRoutingIcon())
        if ServerFeatureEnum.P2P in row_data.features:
            feature_icons.append(P2PIcon())
        if ServerFeatureEnum.TOR in row_data.features:
            feature_icons.append(TORIcon())
        return feature_icons

    @property
    def details_visible(self) -> bool:
        """Returns whether the row has interactive details (connect/upgrade button)."""
        return self._hover_stack.get_hover_child() is not None

    @property
    def action_text(self) -> str:
        """Returns the text of the connect/upgrade action label."""
        return self._hover_action_label.get_text()

    @property
    def label_sensitive(self) -> bool:
        """Returns whether the row is not dimmed."""
        return not self.has_css_class("dimmed")

    @property
    def icon(self) -> Optional[Gtk.Widget]:
        """Returns the leave-child icon widget for this row, or None if not set."""
        return self._icon_leave

    @property
    def connect_button_tooltip(self) -> Optional[str]:
        """Returns the connect button tooltip text."""
        return self.action_button.get_tooltip_text()

    @property
    def server_load(self) -> Optional[str]:
        """Returns the server load label text, or None if not visible."""
        if self._server_load.get_visible():
            return self._server_load.get_label()
        return None

    @GObject.Signal(name="toggle-children")
    def toggle_children(self):
        """Signal emitted when the user clicks the button to expand/collapse child rows."""

    @property
    def label(self):
        """Returns the name of the country this row content is for."""
        return self._label.get_text()

    @property
    def expanded(self):
        """Returns whether the row is expanded showing children rows if any."""
        return self._expanded

    @expanded.setter
    def expanded(self, value: bool):
        """Sets whether children rows should be shown or not."""
        self._expanded = value
        self.toggle_button.set_child(
            self._expanded_img if self.expanded else self._collapsed_img
        )
        tooltip_text = (
            self._row_data.toggle_button_tooltips[1]
            if self.expanded else self._row_data.toggle_button_tooltips[0]
        )
        self.toggle_button.set_tooltip_text(tooltip_text)

    def _set_toggle_button_visible(self, visible: bool):
        """Show or hide the toggle button while keeping it in the layout.

        set_opacity(0) keeps the space reserved to avoid layout shifts,
        but we also set_sensitive(False) so it is removed from the tab
        order and screen reader tree when invisible.
        """
        self.toggle_button.set_opacity(1 if visible else 0)
        self.toggle_button.set_sensitive(visible)

    def _on_toggle_button_clicked(self, _toggle_button: Gtk.Button):
        self.expanded = not self.expanded
        self.emit("toggle-children")

    def _on_connect_button_clicked(self, _connect_button: Gtk.Button):
        self._row_data.on_connect()
        self._hover_stack.show_leave_child()

    def _on_upgrade_button_clicked(self, _connect_button: Gtk.Button):
        Gtk.show_uri(self.get_root(), UPGRADE_URL, 0)
        self._hover_stack.show_leave_child()

    def click_toggle_button(self):
        """Clicks the button to toggle the country servers.
        This method was made available for tests."""
        self.toggle_button.emit("clicked")

    def click_action_button(self):
        """Clicks the button to connect to the country.
        This method was made available for tests."""
        self.action_button.emit("clicked")

    def get_feature_icons(self) -> List[Gtk.Image]:
        """Returns the list of feature icons currently displayed."""
        icons = []
        child = self._feature_icons_box.get_first_child()
        while child:
            icons.append(child)
            child = child.get_next_sibling()
        return icons

    @property
    def server_features(self) -> Set[ServerFeatureEnum]:
        """Returns the set of features supported by the servers in this country."""
        return self._row_data.features

    def grab_focus(self):  # pylint: disable=arguments-differ
        """Focuses on the connect button if available, otherwise the toggle."""
        if not self._row_data.under_maintenance:
            self._hover_stack.show_hover_child()
            self.action_button.grab_focus()
        elif self.toggle_button.get_sensitive():
            self.toggle_button.grab_focus()

    def _on_unrealize(self, _widget):
        """Called when widget is unrealized - performs cleanup."""
        self.reset()

    def reset(self):
        """Resets the state of this row content."""
        self.remove_css_class("dimmed")
        for signal_id, widget in self._connected_signals:
            widget.disconnect(signal_id)
        self._connected_signals.clear()
        self._hover_stack.reset()
        self._remove_accessibility_relations()
        self._remove_icons()
        self._row_data = None

    def _remove_accessibility_relations(self):
        """Clears all accessibility relations from the connect button."""
        remove_accessibility(self.action_button, Gtk.AccessibleRelation.DESCRIBED_BY)
        remove_accessibility(self.action_button, Gtk.AccessibleRelation.LABELLED_BY)

    def _remove_icons(self):
        """Removes dynamically added icon widgets and feature icons."""
        if self._icon_leave and self._icon_leave.get_parent() == self._leave_box:
            self._leave_box.remove(self._icon_leave)
            self._icon_leave = None

        if self._icon_hover and self._icon_hover.get_parent() == self._hover_button_content:
            self._hover_button_content.remove(self._icon_hover)
            self._icon_hover = None

        for icon in self._feature_icons:
            self._feature_icons_box.remove(icon)
        self._feature_icons = []



"""
This module defines the widget used to display the 2FA form.

Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from concurrent.futures import Future
from typing import Optional
from gi.repository import GLib, GObject

from proton.vpn import logging

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.login.two_factor_auth.authenticate_button import AuthenticateButton
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget
from proton.vpn.app.gtk.widgets.main.notifications import Notifications

logger = logging.getLogger(__name__)


class AuthenticatorAppForm(Gtk.Box):  # pylint: disable=too-many-instance-attributes
    """
    Implements the UI for TOTP authentication. Once the right 2FA code
    is provided, it emits the `two-factor-auth-successful` signal.
    """
    TWOFA_HELP_LABEL = "6-digit authentication code"
    TWOFA_ENTRY_PLACEHOLDER = "123456"
    TWOFA_ENTRY_LIMIT_CLARIFICATION = ""

    RECOVERY_HELP_LABEL = "Recovery code"
    RECOVERY_ENTRY_PLACEHOLDER = "y5d6132f"
    RECOVERY_ENTRY_LIMIT_CLARIFICATION = "Each code can only be used once"

    RECOVERY_TOGGLE_AUTHENICATION_MODE_LABEL = "Use 6-digit authentication code"
    TWOFA_TOGGLE_AUTHENICATION_MODE_LABEL = "Use recovery code"

    TWOFA_REQUIRED_CHARACTERS = 6
    RECOVERY_REQUIRED_CHARACTERS = 8

    INCORRECT_TWOFA_CODE_MESSAGE = "Incorrect 2FA code."
    LOGGING_IN_MESSAGE = "Signing in..."

    def __init__(
            self,
            controller: Controller,
            notifications: Notifications,
            overlay_widget: OverlayWidget,
            authenticate_button: Optional[AuthenticateButton] = None
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=30)

        self._controller = controller
        self._notifications = notifications
        self._overlay_widget = overlay_widget

        self.set_name("two-factor-auth-form")
        self._display_2fa_mode = True

        self._entry_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=10)
        self.append(self._entry_box)

        # pylint: disable=R0801
        self._code_entry = Gtk.Entry()
        safe_signal_connect(
            self._code_entry, "changed", self._on_entry_changed
        )
        self._code_entry.set_input_purpose(Gtk.InputPurpose.FREE_FORM)

        self._entry_limit_clarification_label = Gtk.Label()
        self._entry_limit_clarification_label.set_halign(Gtk.Align.START)
        self._entry_limit_clarification_label.add_css_class("dim-label")
        self._entry_limit_clarification_label.set_wrap(True)

        self._help_label = Gtk.Label()
        self._help_label.set_halign(Gtk.Align.START)
        # Pack the help label and the code entry in the entry box
        self._entry_box.append(self._help_label)
        self._entry_box.append(self._code_entry)
        self._entry_box.append(self._entry_limit_clarification_label)

        # Box is used to group the authenticate button and the toggle authentication mode button.
        self._button_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=15)
        self._button_box.set_margin_top(5)
        self._button_box.set_margin_bottom(5)
        self.append(self._button_box)

        self._authenticate_button = authenticate_button or AuthenticateButton()
        safe_signal_connect(
            self._authenticate_button, "clicked", self._on_authenticate_button_clicked
        )
        self._button_box.append(self._authenticate_button)

        # Button used to toggle between 2FA and recovery mode.
        self._toggle_authentication_mode_button = Gtk.Button(label="")
        self._toggle_authentication_mode_button.add_css_class("secondary")
        self._toggle_authentication_mode_button.set_halign(Gtk.Align.FILL)
        self._toggle_authentication_mode_button.set_hexpand(True)
        safe_signal_connect(
            self._toggle_authentication_mode_button,
            "clicked", self._on_toggle_authentication_mode_clicked
        )
        self._button_box.append(self._toggle_authentication_mode_button)

        # Button to cancel 2FA.
        self._cancel_button = Gtk.Button(label="Cancel")
        self._cancel_button.add_css_class("danger")
        safe_signal_connect(self._cancel_button, "clicked", self._on_cancel_button_clicked)
        self._button_box.append(self._cancel_button)

        # Pressing enter on the password entry triggers the clicked event
        # on the login button.
        safe_signal_connect(
            self._code_entry,
            "activate",
            self._on_authenticate_button_clicked
        )
        self._display_2fa_ui()
        self.reset()

    def _on_toggle_authentication_mode_clicked(self, _):
        self._display_2fa_mode = not self._display_2fa_mode

        if self._display_2fa_mode:
            self._display_2fa_ui()
        else:
            self._display_recovery_ui()

        self._code_entry.grab_focus()

    def _display_2fa_ui(self):
        self.code_entry_placeholder = self.TWOFA_ENTRY_PLACEHOLDER
        self.help_label = self.TWOFA_HELP_LABEL
        self.toggle_authentication_mode_button_label = \
            self.TWOFA_TOGGLE_AUTHENICATION_MODE_LABEL
        self.limit_clarification = self.TWOFA_ENTRY_LIMIT_CLARIFICATION

        self._authenticate_button.enable = self.entry_has_required_amount_of_characters

    def _display_recovery_ui(self):
        self.code_entry_placeholder = self.RECOVERY_ENTRY_PLACEHOLDER
        self.help_label = self.RECOVERY_HELP_LABEL
        self.toggle_authentication_mode_button_label = \
            self.RECOVERY_TOGGLE_AUTHENICATION_MODE_LABEL
        self.limit_clarification = self.RECOVERY_ENTRY_LIMIT_CLARIFICATION

        self._authenticate_button.enable = self.entry_has_required_amount_of_characters

    def _on_entry_changed(self, _):
        """Toggles login button state based on username and password lengths."""
        self._authenticate_button.enable = self.entry_has_required_amount_of_characters

    def reset(self):
        """Resets the state of the login/2fa forms."""
        self.two_factor_auth_code = ""
        self._authenticate_button.enable = False
        self._code_entry.grab_focus()

    @property
    def entry_has_required_amount_of_characters(self) -> bool:
        """Returns if the entry has the required amount of characters."""
        return (
            self._display_2fa_mode
            and len(self.code.strip()) == self.TWOFA_REQUIRED_CHARACTERS
            or
            not self._display_2fa_mode
            and len(self.code.strip()) == self.RECOVERY_REQUIRED_CHARACTERS
        )

    @property
    def code(self):
        """Returns the content of `code_entry`"""
        return self._code_entry.get_text()

    @code.setter
    def code(self, newvalue):
        """Sets the content of `code_entry`"""
        return self._code_entry.set_text(newvalue)

    @property
    def two_factor_auth_code(self):
        """Returns the code introduced in the 2FA form."""
        return self._code_entry.get_text()

    @two_factor_auth_code.setter
    def two_factor_auth_code(self, code: str):
        """Sets the code in the 2FA form."""
        self._code_entry.set_text(code)

    @property
    def limit_clarification(self) -> str:
        """Returns the limit clarification text within `limit_clarification`"""
        return self._entry_limit_clarification_label.get_text()

    @limit_clarification.setter
    def limit_clarification(self, newvalue: str):
        """Sets the limit clarification text within `limit_clarification`"""
        self._entry_limit_clarification_label.set_label(newvalue)

    @property
    def code_entry_placeholder(self) -> str:
        """Sets the placeholder text within `code_entry`"""
        return self._code_entry.get_placeholder_text()

    @code_entry_placeholder.setter
    def code_entry_placeholder(self, newvalue: str):
        """Sets the text within `code_entry`"""
        self._code_entry.set_placeholder_text(newvalue)

    @property
    def help_label(self) -> str:
        """Returns text within `help_label`"""
        return self._help_label.get_label()

    @help_label.setter
    def help_label(self, newvalue: str):
        """Sets the label text within `help_label`"""
        self._help_label.set_label(newvalue)

    @property
    def toggle_authentication_mode_button_label(self) -> str:
        """Returns the label text within `toggle_authentication_mode_button`"""
        return self._toggle_authentication_mode_button.get_label()

    @toggle_authentication_mode_button_label.setter
    def toggle_authentication_mode_button_label(self, newvalue):
        """Sets the label text within `toggle_authentication_mode_button`"""
        self._toggle_authentication_mode_button.set_label(newvalue)

    def _on_cancel_button_clicked(self, _):
        self.emit("two-factor-auth-cancelled")
        self.reset()

    def _on_authenticate_button_clicked(self, _):
        """Called when the authenticate button is clicked."""
        logger.info(
            "Clicked on authenticate via authenticator app",
            category="UI", subcategory="LOGIN-2FA", event="CLICK"
        )
        self._overlay_widget.show_message(self.LOGGING_IN_MESSAGE)

        future = self._controller.submit_2fa_code(self.two_factor_auth_code)
        future.add_done_callback(
            lambda future: GLib.idle_add(self._on_2fa_code_submission_result, future)
        )

    def _on_2fa_code_submission_result(self, future: Future):
        try:
            result = future.result()
        finally:
            self.reset()
            self._overlay_widget.hide()

        if result.success:
            self.emit("two-factor-auth-successful")
        else:
            self._notifications.show_error_message(self.INCORRECT_TWOFA_CODE_MESSAGE)
            logger.warning(
                self.INCORRECT_TWOFA_CODE_MESSAGE, category="APP",
                subcategory="LOGIN-2FA", event="RESULT"
            )

    @GObject.Signal
    def two_factor_auth_successful(self):
        """Signal emitted after a successful 2FA."""

    @GObject.Signal
    def two_factor_auth_cancelled(self):
        """Signal emitted after 2FA was cancelled by the user."""

    @GObject.Signal
    def toggle_authentication_mode_button_clicked(self):
        """Signal emitted after the toggle authentication mode button is clicked."""

    def toggle_authentication_button_click(self):
        """Emulates the click of a button.
        This method was made public for testing purposes.
        """
        self._toggle_authentication_mode_button.emit("clicked")

    def authenticate_button_click(self):
        """Submits the 2FA form."""
        self._authenticate_button.emit("clicked")

    @property
    def authenticate_button_enabled(self):
        """Returns if the authenticate button is enabled."""
        return self._authenticate_button.enable



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from types import ModuleType
from typing import Optional, cast

from gi.repository import Gtk

from proton.vpn.core.settings.split_tunneling import SplitTunnelingMode

from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import SettingName
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.app.selected_app_list \
    import SelectedAppList
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.app.app_select_window \
    import AppSelectionWindow
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.app.data_structures \
    import AppData
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.app.installed_apps \
    import get_all_installed_apps


LABEL_CONVERSION = {
    SplitTunnelingMode.INCLUDE: "Included apps ",
    SplitTunnelingMode.EXCLUDE: "Excluded apps ",
}


class AppBasedSplitTunnelingSettings(Gtk.Box):  # pylint: disable=too-many-instance-attributes
    """This class contains all settings and configurations
    related to app based split tunneling, serving as an
    entry point to anything related with app based split tunneling.
    """
    def __init__(
        self,
        controller: Controller,
        mode: SplitTunnelingMode,
        setting_path_name_template: str =
        "settings.features.split_tunneling.[mode].app_paths",
        stored_apps: Optional[list[str]] = None,
        selected_app_list: Optional[SelectedAppList] = None,
        installed_apps: Optional[list[AppData]] = None,
        gtk: ModuleType = Gtk,
    ):  # pylint: disable=too-many-arguments
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=20)
        self.set_name("split-tunneling-app-based-settings")

        self._controller = controller
        self._setting_path_name_template = setting_path_name_template
        self._mode = mode
        self._stored_apps = stored_apps if stored_apps is not None else \
            self._get_settings()
        self._installed_apps = installed_apps if installed_apps is not None else \
            get_all_installed_apps()
        self._selected_app_list = selected_app_list if selected_app_list is not None else\
            SelectedAppList(self._get_selected_app_list())

        self.gtk = gtk

        self._mode_label = SettingName("")
        self._app_count_label = SettingName("")
        self._add_button = self._create_add_button()
        self._add_app_window: Optional[AppSelectionWindow] = None

        mode_and_app_count_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        mode_and_app_count_box.set_halign(Gtk.Align.START)
        mode_and_app_count_box.append(self._mode_label)
        mode_and_app_count_box.append(self._app_count_label)

        self.append(mode_and_app_count_box)
        self.append(self._selected_app_list)
        self.append(self._add_button)

        self._update_mode_label()
        self._update_app_count_label()

        # We need to track whenever an app is removed or added to the list
        safe_signal_connect(self._selected_app_list, "app-removed", self._on_app_removed)
        safe_signal_connect(
            self._selected_app_list, "app-list-refreshed", self._on_app_list_refreshed
        )

    def _create_add_button(self) -> Gtk.Button:
        button = self.gtk.Button.new_with_label("Add")
        button.set_name("split-tunneling-app-add-button")
        button.add_css_class("secondary")
        safe_signal_connect(button, "clicked", self._on_clicked_add)
        button.set_hexpand(True)
        button.set_halign(Gtk.Align.START)

        return button

    def _on_clicked_add(self, _: Gtk.Button):
        self._add_app_window = AppSelectionWindow(
            title=self._window_title,
            controller=self._controller,
            stored_apps=self._stored_apps,
            installed_apps=self._installed_apps
        )

        safe_signal_connect(
            self._add_app_window, "app-selection-completed", self._on_app_selection_completed
        )
        safe_signal_connect(
            self._add_app_window, "unrealize", self._on_add_app_window_unrealize
        )
        self._add_app_window.present()

    def _on_add_app_window_unrealize(self, _):
        self._add_app_window = None

    @property
    def _window_title(self) -> str:
        return f"Add {LABEL_CONVERSION[self._mode].lower()}"

    def _on_app_selection_completed(
        self, _: AppSelectionWindow,
        selected_apps: list[AppData]
    ):
        """Receive the newly selected apps and issue a refresh to the list
        with the new apps.

        Args:
            _ (AppSelectionWindow): Discarded.
            selected_apps (list[AppData]): List with selected apps to split tunnel.
        """
        # This emits the app-list-refreshed signal
        self._selected_app_list.refresh(selected_apps)

    def update_list_on_new_mode(self, mode: SplitTunnelingMode):
        """Updates the list of split tunneled apps when the mode is switched.

        Args:
            mode (SplitTunnelingMode): the mode that we need to switch to
        """
        self._mode = mode
        self._stored_apps = self._get_settings()
        selected_app_list = self._get_selected_app_list()
        self._selected_app_list.refresh(selected_app_list)
        self._update_mode_label()

    def _on_app_removed(self, _: SelectedAppList, app_data: AppData):
        self._stored_apps = [
            app_exec for app_exec in self._stored_apps
            if app_exec != app_data.executable
        ]
        self._save_and_update_app_count()

    def _on_app_list_refreshed(self, _: SelectedAppList, selected_app_data: list[AppData]):
        self._stored_apps = [app_data.executable for app_data in selected_app_data]
        self._save_and_update_app_count()

    def _save_and_update_app_count(self):
        self._save_settings()
        self._update_app_count_label()

    def _update_app_count_label(self):
        self._app_count_label.set_label(f"({self.amount_of_selected_apps})")

    def _update_mode_label(self):
        self._mode_label.set_label(f"{LABEL_CONVERSION[self._mode]}")

    def _get_selected_app_list(self) -> list[str]:
        return [
            app
            for stored_app_path in self._stored_apps
            for app in self._installed_apps
            if stored_app_path == app.executable
        ]

    def _get_settings(self) -> list[str]:
        return cast(list[str], self._controller.get_setting_attr(self._setting_path_name))

    def _save_settings(self):
        self._controller.save_setting_attr(self._setting_path_name, self._stored_apps)

    @property
    def _setting_path_name(self) -> str:
        return self._setting_path_name_template.replace("[mode]", self._mode.value)

    def get_app_count_label(self) -> str:
        """Returns the label that holds the app counter.

        Returns:
            str:
        """
        return self._app_count_label.get_label()

    def get_mode_label(self) -> str:
        """Returns the label that holds the mode.

        Returns:
            str:
        """
        return self._mode_label.get_label()

    @property
    def amount_of_selected_apps(self) -> int:
        """Returns the amount of selected apps in the list.

        Returns:
            int:
        """
        return len(self._stored_apps)

    def emit_signal_app_removed(self, app_data: AppData):
        """Emits the app-removed signal to the selected app list.
        Mainly used for testing purposes.

        Args:
            app_data (AppData): The app data that was removed.
        """
        self._selected_app_list.emit("app-removed", app_data)

    def emit_signal_app_list_refreshed(self, app_data_list: list[AppData]):
        """Emits the app-list-refreshed signal to the selected app list.
        Mainly used for testing purposes.

        Args:
            app_data_list (list[AppData]): The list of app data that was refreshed.
        """
        self._selected_app_list.emit("app-list-refreshed", app_data_list)

    def click_on_add_button(self):
        """Clicks on the add button."""
        self._add_button.clicked()



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from dataclasses import dataclass
from typing import ClassVar, List

from gi.repository import Gtk, GObject

from proton.vpn.core.settings.split_tunneling import SplitTunnelingMode
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import \
    SettingName, SettingDescription


@dataclass
class ModeData:
    """Dataclass that contains display mode information."""
    title: str
    description: str
    mode: SplitTunnelingMode


EXCLUDE_MODE = ModeData(
    title="Exclude mode",
    description="Allow selected apps to connect "
    "without VPN protection.",
    mode=SplitTunnelingMode.EXCLUDE
)

INCLUDE_MODE = ModeData(
    title="Include mode",
    description="Only selected apps connect with "
    "VPN protection. All other traffic is unprotected.",
    mode=SplitTunnelingMode.INCLUDE,
)


class ModeRadioButton(Gtk.Box):
    """Custom radio button for mode selection"""
    # Class-level group storage
    _radio_buttons: ClassVar[List[Gtk.CheckButton]] = []

    def __init__(self, mode_data: ModeData):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)

        self.mode = mode_data.mode

        self._radio_button = Gtk.CheckButton()

        if not self._radio_buttons:
            self._radio_buttons.append(self._radio_button)
        else:
            self._radio_button.set_group(self._radio_buttons[0])

        self._radio_button.set_halign(Gtk.Align.START)
        self._radio_button.set_valign(Gtk.Align.START)
        self._radio_button.set_margin_top(2)

        text_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=4)

        self._title_label = SettingName(mode_data.title, bold=True)

        self._description_label = SettingDescription(mode_data.description)

        text_box.append(self._title_label)
        text_box.append(self._description_label)

        self.append(self._radio_button)
        self.append(text_box)

        self.add_css_class("setting-item")
        self.set_margin_top(8)
        self.set_margin_bottom(8)

        safe_signal_connect(self._radio_button, "toggled", self._on_radio_toggled)

    def _on_radio_toggled(self, _):
        self.emit("toggled", self.mode)

    @GObject.Signal(name="toggled", arg_types=(object,))
    def toggled(self, widget: SplitTunnelingMode):
        """Signal that the radio button was toggled.

        Args:
            widget (SplitTunnelingMode): widget that was toggled.
        """

    def get_active(self) -> bool:
        """Get the active state of the radio button"""
        return self._radio_button.get_active()

    def set_active(self, newvalue: bool):
        """Set the active state of the radio button"""
        self._radio_button.set_active(newvalue)


SETTINGS_PATH_NAME = "settings.features.split_tunneling.mode"


class SplitTunnelingModeSetting(Gtk.Box):
    """Object for building UI based on excluding or including split tunneling.
    """
    def __init__(
            self,
            controller: Controller,
            setting_path_name: str = SETTINGS_PATH_NAME,
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        self._controller = controller
        self._setting_path_name = setting_path_name

        self._exclude_radio = None
        self._include_radio = None
        self.mode = self._get_setting()
        self.build()

    def build(self):
        """Used to build UI.
        """
        self._exclude_radio = ModeRadioButton(mode_data=EXCLUDE_MODE)
        self._include_radio = ModeRadioButton(mode_data=INCLUDE_MODE)

        self.append(self._exclude_radio)
        self.append(self._include_radio)

        self._update_selection()

        safe_signal_connect(self._exclude_radio, "toggled", self._on_mode_changed)
        safe_signal_connect(self._include_radio, "toggled", self._on_mode_changed)

    @GObject.Signal(name="mode-switched", arg_types=(object,))
    def mode_switched(self, mode_changed: SplitTunnelingMode):
        """When the toggle is switched, the mode it switched to is sent
        to the subscribers.

        Args:
            mode_changed (SplitTunnelingMode): mode that was changed to.
        """

    def _on_mode_changed(self, radio_button: ModeRadioButton, mode: SplitTunnelingMode):
        """Handle mode selection change"""
        if radio_button.get_active():
            self._save_setting(mode)
            self.mode = mode
            self.emit("mode-switched", mode)

    def _update_selection(self):
        """Update radio button selection based on current setting"""
        current_mode_exclude = self.mode == SplitTunnelingMode.EXCLUDE
        self._exclude_radio.set_active(current_mode_exclude)
        self._include_radio.set_active(not current_mode_exclude)

    def _get_setting(self) -> SplitTunnelingMode:
        """Get current mode setting"""
        return self._controller.get_setting_attr(self._setting_path_name)

    def _save_setting(self, mode: SplitTunnelingMode):
        """Save mode setting"""
        self._controller.save_setting_attr(self._setting_path_name, mode)

    def get_exclude_radio_button(self):
        """Get the exclude radio button."""
        return self._exclude_radio

    def get_include_radio_button(self):
        """Get the include radio button."""
        return self._include_radio



"""
GLib utility functions.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from concurrent.futures import Future
from typing import Callable

from gi.repository import GLib


def run_once(function: Callable, *args, priority=GLib.PRIORITY_DEFAULT, **kwargs) -> int:
    """
    Python implementation of GLib.idle_add_once, as currently is not available
    in pygobject.
    https://docs.gtk.org/glib/func.idle_add_once.html.
    """
    def wrapper_function():
        function(*args, **kwargs)
        # Returning a falsy value is required so that GLib does not keep
        # running the function over and over again.
        return False

    return GLib.idle_add(wrapper_function, priority=priority)


def run_periodically(function, *args, interval_ms: int, **kwargs) -> int:
    """
    Runs a function periodically on the GLib main loop.

    :param function: function to be called periodically
    :param *args: arguments to be passed to the function.
    :param interval_ms: interval at which the function should be called.
    :param **kwargs: keyword arguments to be passed to the function.
    """
    run_once(function, *args, **kwargs)

    def wrapper_function():
        function(*args, **kwargs)
        # True is returned so that GLib keeps running the function.
        return True

    return GLib.timeout_add(interval_ms, wrapper_function)


def bubble_up_errors(future: Future):
    """Makes sure that any error the future resolves to bubbles up to the GLib main loop."""
    future.add_done_callback(lambda f: GLib.idle_add(f.result))


def add_done_callback(future: Future, callback: Callable[[Future], None]):
    """
    Adds a callback to be executed once the future completes, but in a way that's GLib-compatible.

    We currently use Futures to bridge between the asyncio event loop and the GLib event loop.
    When using Future.add_done_callback(callback), the callback is run by the thread
    running the asyncio loop. However, if the callback does GTK-related work, it needs to
    run on the thread running the GLib loop.

    This function takes care of wrapping the callback in a GLib.idle_add call to make sure the
    callback runs on the GLib thread.
    """
    future.add_done_callback(lambda f: GLib.idle_add(callback, f))



"""
This module defines the location rows displayed in the server list widget.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from itertools import chain
from typing import List, Optional

from gi.repository import GLib

from proton.vpn.session.servers import Location, TierEnum
from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.expandable_row import ExpandableRow
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.row_content import RowContent
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.row_view_model import RowViewModel
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.utils import (
    make_connect_callback, sync_rows_with_model_items
)
from proton.vpn.app.gtk.widgets.vpn.serverlist.icons import LocationIcon


class LocationRow(Gtk.Box):
    """Row representing a location in the server list widget."""

    def __init__(self):
        # pylint: disable=duplicate-code
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self._location: Optional[Location] = None
        self._controller = None
        self._user_tier = None
        self._expandable_row = ExpandableRow(
            on_expand=self._add_server_rows,
            on_collapse=self._remove_server_rows,
        )
        self.append(self._expandable_row)
        self._server_rows: List[RowContent] = []

    # pylint: disable=too-many-arguments
    def display(
        self, controller: Controller, location: Location, user_tier: int,
        expanded: bool = False
    ):
        """Displays the location row according to the specified parameters.

        Args:
            controller: The controller instance
            location: The location to display
            user_tier: The user's tier level
            expanded: Whether the location row should be expanded (defaults to False)
        """
        self.reset(keep_server_rows=expanded)
        self._controller = controller
        self._location = location
        self._user_tier = user_tier
        self._expandable_row.connect_toggle()
        upgrade_required = user_tier == TierEnum.FREE and not location.free

        row_data = RowViewModel(
            name=location.name,
            on_connect=make_connect_callback(controller, location.servers, user_tier),
            free=location.free,
            under_maintenance=location.under_maintenance and not upgrade_required,
            features=location.features,
            smart_routing=location.smart_routing,
            toggable=True,
            upgrade_required=upgrade_required,
            icon_factory=LocationIcon,
            connect_button_tooltip=(
                f"Upgrade to connect to {location.name}"
                if upgrade_required else
                f"Connect to {location.name}"
            ),
            toggle_button_tooltips=(
                f"Show all servers from {location.name}",
                f"Hide all servers from {location.name}",
            ),
        )
        self._expandable_row.row_content.display(row_data)
        if expanded:
            self.click_toggle_button()

    def reset(self, keep_server_rows: bool = False):
        """Resets the location row to its initial state."""
        self._expandable_row.reset(keep_children=keep_server_rows)

    @property
    def label(self) -> str:
        """Returns the location label."""
        return self._expandable_row.row_content.label

    @property
    def server_rows(self) -> List[RowContent]:
        """Returns the list of server rows currently displayed."""
        return list(self._server_rows)

    @property
    def expanded(self) -> bool:
        """Returns whether the location row is currently expanded or not."""
        return self._expandable_row.row_content.expanded

    def grab_focus(self):  # pylint: disable=arguments-differ
        """See Gtk.Widget.grab_focus()"""
        self._expandable_row.row_content.grab_focus()

    def click_toggle_button(self):
        """Simulates a click on the toggle button to expand/collapse the row."""
        self._expandable_row.row_content.click_toggle_button()

    def _remove_server_rows(self):
        while self._server_rows:
            server_row = self._server_rows.pop()
            self._expandable_row.remove_child(server_row)
            server_row.reset()

    def _add_server_rows(self):
        servers = self._location.servers
        if self._user_tier == TierEnum.FREE and self._location.free:
            servers = chain(self._location.free_servers, self._location.paid_servers)
        else:
            servers = chain(self._location.paid_servers, self._location.free_servers)

        # Capture controller directly to avoid closing over `self` in on_connect
        controller = self._controller

        # pylint: disable=duplicate-code
        def display_server_row(server_row, server):
            upgrade_required = self._user_tier == TierEnum.FREE and not server.free

            def on_connect():
                future = controller.connect_to_server(server.name)
                future.add_done_callback(lambda f: GLib.idle_add(f.result))

            row_data = RowViewModel(
                name=server.name,
                on_connect=on_connect,
                free=server.free,
                under_maintenance=server.under_maintenance and not upgrade_required,
                features=server.features,
                smart_routing=server.smart_routing,
                toggable=False,
                upgrade_required=upgrade_required,
                load=None if server.under_maintenance else server.load,
                connect_button_tooltip=(
                    f"Upgrade to connect to {server.name}"
                    if upgrade_required else
                    f"Connect to {server.name}"
                ),
            )
            server_row.display(row_data)

        sync_rows_with_model_items(
            list(servers),
            self._server_rows,
            self._expandable_row.container,
            RowContent,
            display_server_row
        )



"""
This module defines the widget used to display the 2FA methods

Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Optional
from gi.repository import Gtk, GObject

from proton.vpn import logging

from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.login.logo import TwoFactorAuthProtonVPNLogo
from proton.vpn.app.gtk.widgets.main.notifications import Notifications
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget
from proton.vpn.app.gtk.widgets.login.two_factor_auth.two_factor_auth_stack import\
    TwoFactorAuthStack

logger = logging.getLogger(__name__)


class TwoFactorAuthWidget(Gtk.Box):
    """Widget used to display the 2FA methods."""
    TWO_FACTOR_AUTH_LABEL = "Two-factor authentication"

    def __init__(
        self,
        controller: Controller,
        notifications: Notifications,
        overlay_widget: OverlayWidget,
        two_factor_auth_stack: Optional[TwoFactorAuthStack] = None
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=20)
        self.set_name("two-factor-auth-widget")

        self._two_factor_auth_title = Gtk.Label(label=self.TWO_FACTOR_AUTH_LABEL)
        self._two_factor_auth_title.set_halign(Gtk.Align.CENTER)
        self._two_factor_auth_title.add_css_class("two-factor-auth-stack-title")

        self.two_factor_auth_stack = two_factor_auth_stack \
            or TwoFactorAuthStack(controller, notifications, overlay_widget)

        self.stack_switch = Gtk.StackSwitcher()
        self.stack_switch.set_hexpand(True)
        self.stack_switch.set_name("stack-switch")
        self.stack_switch.set_halign(Gtk.Align.FILL)
        self.stack_switch.set_valign(Gtk.Align.CENTER)

        self.stack_switch.set_stack(self.two_factor_auth_stack)

        # Ensure the children of the stack switcher fill the width of the stack switcher
        child = self.stack_switch.get_first_child()
        while child:
            child.set_hexpand(True)
            child.set_halign(Gtk.Align.FILL)
            child = child.get_next_sibling()

        self.append(TwoFactorAuthProtonVPNLogo())
        self.append(self._two_factor_auth_title)
        self.append(self.stack_switch)
        self.append(self.two_factor_auth_stack)

        safe_signal_connect(
            self.two_factor_auth_stack,
            "two-factor-auth-successful",
            self._on_two_factor_auth_successful
        )
        safe_signal_connect(
            self.two_factor_auth_stack,
            "two-factor-auth-cancelled",
            self._on_two_factor_auth_cancelled
        )

    def _on_two_factor_auth_successful(self, _):
        self.emit("two-factor-auth-successful")

    def _on_two_factor_auth_cancelled(self, _):
        self.emit("two-factor-auth-cancelled")

    def reset(self):
        """Resets the widget to its initial state."""
        self.two_factor_auth_stack.reset()

    @GObject.Signal
    def two_factor_auth_successful(self):
        """Signal emitted after a successful 2FA."""

    @GObject.Signal
    def two_factor_auth_cancelled(self):
        """Signal emitted after 2FA was cancelled by the user."""



"""
Server search entry module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from gi.repository import GObject

from proton.vpn import logging

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect

logger = logging.getLogger(__name__)


class SearchEntry(Gtk.SearchEntry):
    """Widget used to filter server list based on user input."""
    def __init__(self):
        super().__init__()
        self.set_name("search-entry")
        self.props.placeholder_text = "Press Ctrl+F to search"  # pylint: disable=no-member
        safe_signal_connect(self, "request-focus", self.request_focus)

    def _on_request_focus(self, _):
        self.grab_focus()

    @GObject.Signal(name="request_focus", flags=GObject.SignalFlags.ACTION)
    def request_focus(self, _):
        """Emitting this signal requests input focus on the search text entry."""

    def reset(self):
        """Resets the widget UI."""
        self.set_text("")



"""
Runtime assertion utility.

Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""


def runtime_assert(condition: bool, message: str = "") -> None:
    """Raises RuntimeError if condition is False.

    Unlike assert statements, this is never stripped by the Python optimizer (-O flag)
    and does not trigger Bandit B101 warnings.
    """
    if not condition:
        raise RuntimeError(message)



"""
This module defines the main application window.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from concurrent.futures import Future
import logging
from typing import Optional

from gi.repository import GLib, Gtk

from proton.session.exceptions import ProtonAPINotReachable, ProtonAPIError
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.exceptions import NPSError
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.main_widget import MainWidget
from proton.vpn.app.gtk.widgets.headerbar.headerbar import HeaderBar
from proton.vpn.app.gtk.widgets.main.notification_bar import NotificationBar
from proton.vpn.app.gtk.widgets.main.notifications import Notifications
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget
from proton.vpn.app.gtk.widgets.main.pull_notifications.nps_survey_modal import \
    NPSSurveyModal
from proton.vpn.session.dataclasses import NPSSurveyResponse


logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class MainWindow(Gtk.ApplicationWindow):
    """Main window."""

    WIDTH = 450
    HEIGHT = 700

    # pylint: disable=too-many-arguments
    def __init__(
            self, application: Gtk.Application,
            controller: Controller,
            notifications: Optional[Notifications] = None,
            header_bar: Optional[HeaderBar] = None,
            main_widget: Optional[MainWidget] = None,
            overlay_widget: Optional[OverlayWidget] = None
    ):
        super().__init__(application=application)
        self._application = application
        self.get_settings().props.gtk_application_prefer_dark_theme = True
        self._controller = controller
        self._close_window_handler_id: Optional[int] = None
        self._shortcut_controller: Optional[Gtk.ShortcutController] = None
        self._nps_modal: Optional[NPSSurveyModal] = None

        self._configure_window()

        self._overlay_widget = overlay_widget or OverlayWidget()

        notifications = notifications or Notifications(
            main_window=self, notification_bar=NotificationBar()
        )

        self.header_bar = header_bar or HeaderBar(
            controller=controller,
            main_window=self,
            overlay_widget=self._overlay_widget
        )
        self.set_titlebar(self.header_bar)
        self.set_title("Proton VPN")

        self.main_widget = main_widget or MainWidget(
            controller=controller,
            main_window=self,
            notifications=notifications,
            overlay_widget=self._overlay_widget
        )
        self.set_child(self.main_widget)

        safe_signal_connect(self, "notify::visible", self._display_pending_notifications)

        self.main_widget.set_visible(True)

    @property
    def application(self) -> Gtk.Application:
        """Returns Gtk.Application object which contains references to windows,
        tray indicator and other settings."""
        return self._application

    def add_keyboard_shortcut(self, target_widget: Gtk.Widget, target_signal: str, shortcut: str):
        """
        Adds a keyboard shortcut so that when pressed it causes the target signal
        to be triggered on the target widget.

        :param target_widget: The widget the keyboard shortcut will trigger the signal on.
        :param target_signal: The signal the keyboard shortcut will trigger on the target widget.
        :param shortcut: The keyboard shortcut string (e.g. "<Ctrl>q")
        """
        def callback(*_):
            target_widget.emit(target_signal)
            return True

        shortcut_trigger = Gtk.ShortcutTrigger.parse_string(shortcut)
        if shortcut_trigger is None:
            raise ValueError(f"Invalid shortcut: {shortcut}")

        shortcut_action = Gtk.CallbackAction.new(callback)
        shortcut_obj = Gtk.Shortcut.new(shortcut_trigger, shortcut_action)

        if self._shortcut_controller is None:
            self._shortcut_controller = Gtk.ShortcutController()
            self._shortcut_controller.set_scope(Gtk.ShortcutScope.LOCAL)
            self.add_controller(self._shortcut_controller)

        self._shortcut_controller.add_shortcut(shortcut_obj)

    def _configure_window(self):
        """
        Handle delete-event, set window resize restrictions...
        """
        self.set_name("main-window")
        self.set_resizable(False)
        self.set_size_request(MainWindow.WIDTH, MainWindow.HEIGHT)

    def configure_close_button_behaviour(self, tray_indicator_enabled: bool):
        """Configures the behaviour of the button to close the window
        (the x button), depending on if the tray indicator is used or not."""
        if tray_indicator_enabled:
            self._close_window_handler_id = self.configure_close_button_to_hide_window()
        else:
            self._close_window_handler_id = self.configure_close_button_to_trigger_quit_menu_entry()

    def configure_close_button_to_hide_window(self):
        """Configures the x (close window) button so that when clicked,
        the window is hidden instead closed."""
        # Handle the event emitted when the user tries to close the window.
        return safe_signal_connect(
            self,
            "close-request",
            self._on_close_button_clicked_then_hide_window
        )

    def _on_close_button_clicked_then_hide_window(self, *_) -> bool:
        """
        Instead of letting the window x button close the app, therefore
        quitting the app, the action is delegated to the Exit entry in
        the menu bar widget.
        """
        self.set_visible(False)

        # Returning True when handling the close-request stops other handlers
        # from being invoked for this event, therefore preventing the default
        # behaviour:
        # https://docs.gtk.org/gtk4/signal.Window.close-request.html
        return True

    def quit(self):
        """Closes the main window, which quits the app."""
        if self._close_window_handler_id:
            self.disconnect(self._close_window_handler_id)

        self.close()

    def configure_close_button_to_trigger_quit_menu_entry(self):
        """Configures the x (close window) button so that when clicked,
        the Exit menu entry is triggered instead."""
        # Handle the event emitted when the user tries to close the window.
        return safe_signal_connect(
            self,
            "close-request",
            self._on_close_button_clicked_then_click_quit_menu_entry
        )

    def _on_close_button_clicked_then_click_quit_menu_entry(self, *_) -> bool:
        """
        Instead of letting the x button close the app, therefore
        quitting the app, the action is delegated to the Exit entry in
        the menu bar widget, which may request confirmation to the user.
        """
        self.header_bar.menu.quit_button_click()

        # Returning True when handling the close-request stops other handlers
        # from being invoked for this event, therefore preventing the default
        # behaviour:
        # https://docs.gtk.org/gtk4/signal.Window.close-request.html
        return True

    def _display_pending_notifications(self, *_):
        if not self._controller.user_logged_in:
            # need to be logged in to submit NPS Survey response
            return

        if not self.get_visible():
            # ensure we're visible, and not going invisible
            return

        nps_notifications = self._controller.notifications.get_nps_survey_notifications()
        while nps_notifications:
            nps_survey = nps_notifications.pop()
            if not nps_survey.seen and nps_survey.is_active:
                self._controller.set_notification_seen(nps_survey.survey_id)
                GLib.idle_add(self._show_nps_survey)
                break

    def create_nps_survey_modal(self) -> NPSSurveyModal:
        """Creates the NPS survey modal."""
        def submit_nps_survey_feedback(score: int, comments: str):
            nps_user_response = NPSSurveyResponse(
                user_score=score,
                user_comments=comments,
                response_type=NPSSurveyResponse.ResponseType.SUBMIT
            )
            future = self._controller.submit_nps_survey_response(nps_user_response)
            future.add_done_callback(self._on_nps_submission_result)

        def dismiss_nps_survey():
            nps_user_response = \
                NPSSurveyResponse(response_type=NPSSurveyResponse.ResponseType.DISMISS)
            future = self._controller.submit_nps_survey_response(nps_user_response)
            future.add_done_callback(self._on_nps_submission_result)

        return NPSSurveyModal(
            self._controller,
            submit_handler=submit_nps_survey_feedback,
            dismiss_handler=dismiss_nps_survey
        )

    def _show_nps_survey(self):
        self._nps_modal = self.create_nps_survey_modal()
        self._nps_modal.set_transient_for(self)
        safe_signal_connect(self._nps_modal, "unrealize", self._on_nps_modal_unrealize)
        self._nps_modal.show()
        return GLib.SOURCE_REMOVE

    def _on_nps_modal_unrealize(self, _):
        self._nps_modal = None

    def _on_nps_submission_result(self, future: Future):
        try:
            future.result()
        except ProtonAPINotReachable:
            logger.warning("NPS survey submission failed: API not reachable.")
        except ProtonAPIError as exc:
            logger.warning("Proton API error: %s", exc)
        except Exception as exc:  # pylint: disable=broad-except
            logger.exception("Unexpected error submitting NPS survey response.")

            def _reraise_on_main_thread(exc=exc):
                raise NPSError(
                    "Unexpected error submitting NPS survey response."
                ) from exc

            GLib.idle_add(_reraise_on_main_thread)



"""
Reusable expandable row with header and revealable content.

Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from typing import Callable, List, Tuple

from proton.vpn.app.gtk import Gtk

from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.row_content import RowContent
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.utils import get_children


class ExpandableRow(Gtk.Box):
    """Reusable expandable row with header and revealable content area."""

    def __init__(
        self,
        on_expand: Callable[[], None],
        on_collapse: Callable[[], None],
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self._on_expand = on_expand
        self._on_collapse = on_collapse
        self._row_content = RowContent()
        self.append(self._row_content)
        self._revealer = Gtk.Revealer()
        self._container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self._container.set_spacing(5)
        self._container.set_margin_top(5)
        self._revealer.set_child(self._container)
        self.append(self._revealer)
        self._connected_signals: List[Tuple[int, Gtk.Widget]] = []
        self._revealer_collapse_signal_id = None
        safe_signal_connect(self, "unrealize", self._on_unrealize)

    @property
    def row_content(self) -> RowContent:
        """The row content (header area)."""
        return self._row_content

    @property
    def container(self) -> Gtk.Box:
        """The container for revealed content. Add child widgets here."""
        return self._container

    def get_children(self) -> list:
        """Returns the list of children in the content container."""
        return get_children(self._container)

    def remove_child(self, child: Gtk.Widget) -> None:
        """Removes a child from the content container."""
        self._container.remove(child)

    def connect_toggle(self) -> None:
        """Connects the toggle-children signal. Call after displaying the row content."""
        signal_id = safe_signal_connect(self._row_content, "toggle-children", self._on_toggle)
        self._connected_signals.append((signal_id, self._row_content))

    def _on_toggle(self, row_content: RowContent) -> None:
        self._set_revealed(row_content.expanded)

    def _set_revealed(self, expanded: bool) -> None:
        if expanded:
            if self._on_expand is not None:
                self._on_expand()
        else:
            self._schedule_collapse_on_reveal_complete()
        self._revealer.set_reveal_child(expanded)

    def set_expanded(self, expanded: bool) -> None:
        """Programmatically set expanded state and run expand/collapse logic."""
        self._row_content.expanded = expanded
        self._set_revealed(expanded)

    def _on_collapse_complete(self, *_args) -> None:
        self._revealer.disconnect(self._revealer_collapse_signal_id)
        self._connected_signals.remove((self._revealer_collapse_signal_id, self._revealer))
        if self._on_collapse is not None:
            self._on_collapse()

    def _schedule_collapse_on_reveal_complete(self) -> None:
        """Waits for the revealer to finish collapsing, then calls _on_collapse."""
        self._revealer_collapse_signal_id = safe_signal_connect(
            self._revealer,
            "notify::child-revealed",
            self._on_collapse_complete
        )
        self._connected_signals.append((self._revealer_collapse_signal_id, self._revealer))

    def _on_unrealize(self, _widget: Gtk.Widget) -> None:
        self.reset()
        self._on_expand = None
        self._on_collapse = None

    def reset(self, keep_children: bool = False) -> None:
        """Resets the row. Disconnects signals and optionally removes children."""
        for signal_id, widget in self._connected_signals:
            widget.disconnect(signal_id)
        self._connected_signals.clear()
        self._row_content.reset()
        if not keep_children and self._on_collapse is not None:
            self._on_collapse()



"""
Login session service — wraps logind dbus interaction.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Any, Callable, Optional

import dbus
from dbus import SystemBus
from dbus.mainloop.glib import DBusGMainLoop

DBusGMainLoop(set_as_default=True)

BUS_NAME = "org.freedesktop.login1"
SEAT_AUTO_PATH = "/org/freedesktop/login1/seat/auto"
SESSION_INTERFACE = "org.freedesktop.login1.Session"
SEAT_INTERFACE = "org.freedesktop.login1.Seat"
PROPERTIES_INTERFACE = "org.freedesktop.DBus.Properties"
UNLOCK_SIGNAL = "Unlock"


class LoginSessionService:
    """
    Provides access to the current login session state via logind over dbus.

    Raises DBusUnavailableError when logind is inaccessible (e.g. under strict
    snap confinement with AppArmor blocking dbus access).
    """

    class DBusUnavailableError(Exception):
        """Raised when logind is inaccessible over dbus."""

    def __init__(
        self,
        bus: SystemBus = None,
        session_object_path: Optional[str] = None,
        dbus_interface: Optional[Callable[[Any, str], Any]] = None
    ):
        self._bus = bus or SystemBus()
        self._session_object_path = session_object_path
        self._dbus_interface = dbus_interface or dbus.Interface

    def add_session_unlocked_signal_receiver(self, handler: Callable) -> object:
        """
        Registers a handler to be called when the session is unlocked.
        Returns a signal receiver object that can be used to remove the handler.
        Raises DBusUnavailableError if logind is inaccessible.
        """
        try:
            if not self._session_object_path:
                self._setup()
        except dbus.exceptions.DBusException as exc:
            raise LoginSessionService.DBusUnavailableError() from exc

        return self._bus.add_signal_receiver(
            handler_function=handler,
            signal_name=UNLOCK_SIGNAL,
            dbus_interface=SESSION_INTERFACE,
            bus_name=BUS_NAME,
            path=self._session_object_path,
        )

    @property
    def is_session_unlocked(self) -> bool:
        """
        Returns True if the session is unlocked, False if locked.
        Raises DBusUnavailableError if logind is inaccessible.
        """
        try:
            if not self._session_object_path:
                self._setup()
            active_session = self._bus.get_object(BUS_NAME, self._session_object_path)
            active_session_properties = self._dbus_interface(active_session, PROPERTIES_INTERFACE)
            return not bool(active_session_properties.Get(SESSION_INTERFACE, "LockedHint"))
        except dbus.exceptions.DBusException as exc:
            raise LoginSessionService.DBusUnavailableError() from exc

    def _setup(self):
        seat_auto_proxy = self._bus.get_object(BUS_NAME, SEAT_AUTO_PATH)
        seat_auto_properties_proxy = self._dbus_interface(seat_auto_proxy, PROPERTIES_INTERFACE)

        seat_properties = seat_auto_properties_proxy.GetAll(SEAT_INTERFACE)
        active_sessions = seat_properties.get("ActiveSession", [])

        if not active_sessions:
            raise RuntimeError("There are no active sessions for this seat")

        _session_id, self._session_object_path = active_sessions



"""
GTK4 Tray Icon Module - Easy to integrate StatusNotifierItem + DBusMenu

This module provides a simple API to add a system tray icon with context menu
to any GTK4 application with minimal code changes.

License: MIT
Author: Matteo Benedetto <me@enne2.net>
Copyright © 2025 Matteo Benedetto

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Usage:
    from trayer import TrayIcon

    # In your GTK4 app:
    tray = TrayIcon(
        app_id="com.example.myapp",
        title="My App",
        icon_name="application-x-executable"
    )

    # Add menu items
    tray.add_menu_item("Show", callback=show_window)
    tray.add_menu_item("Quit", callback=quit_app)

    # Setup (call before app.run())
    tray.setup()
"""
from typing import Callable, Optional
from enum import Enum
from dataclasses import dataclass
import os
import dbus
import dbus.service
import dbus.mainloop.glib
from gi.repository import GLib
from proton.vpn.app.gtk.util import APPLICATION_ID
from proton.vpn import logging

logger = logging.getLogger(__name__)

# Tray icon configuration
TRAY_TITLE = "Proton VPN"
TRAY_ICON_NAME = "proton-vpn-sign"
TRAY_ICON_DESCRIPTION = ""

# StatusNotifierItem specification
SNI_INTERFACE = "org.kde.StatusNotifierItem"
SNI_PATH = "/StatusNotifierItem"

# DBusMenu specification
DBUSMENU_INTERFACE = "com.canonical.dbusmenu"
DBUSMENU_PATH = "/MenuBar"

# StatusNotifierWatcher
SNW_BUS_NAME = "org.kde.StatusNotifierWatcher"
SNW_OBJECT_PATH = "/StatusNotifierWatcher"
SNW_INTERFACE = "org.kde.StatusNotifierWatcher"


class StatusNotifierItemProperty(Enum):
    """Contains the type of data that a status notifier can have."""
    STATUS = "Status"
    CATEGORY = "Category"
    ID = "Id"
    TITLE = "Title"
    ICON_NAME = "IconName"
    MENU = "Menu"
    ITEM_IS_MENU = "ItemIsMenu"
    ICON_ACCESSIBLE_DESCRIPTION = "IconAccessibleDesc"


class MenuType(Enum):
    """Enum for menu types that are used in the menu."""
    ITEM = "standard"
    SEPARATOR = "separator"


@dataclass
class MenuObject:
    """Object that represents an entry in tray menu."""
    id: int  # pylint: disable=invalid-name
    type: MenuType
    label: Optional[str] = None
    callback: Optional[Callable] = None
    enabled: bool = True
    visible: bool = True

    def to_dbus(self) -> dbus.Dictionary:
        """Converts and returns the object to dbus friendly format."""
        return dbus.Dictionary({
            "type": dbus.String(self.type.value),
            "label": dbus.String(self.label),
            "enabled": dbus.Boolean(self.enabled),
            "visible": dbus.Boolean(self.visible)
        }, signature="sv")


class _DBusMenuService(dbus.service.Object):
    """Internal DBusMenu implementation

    This export the objects across Dbus
    https://dbus.freedesktop.org/doc/dbus-python/dbus.service.html?highlight=dbus%20service#dbus.service.Object
    """

    def __init__(self, bus, menu_items: list[MenuObject]):
        self.bus = bus
        self.menu_items = menu_items
        self.revision = 0
        super().__init__(bus, DBUSMENU_PATH)

    def _build_menu_structure(self):
        """Build menu structure from menu_items list"""

        structure = {
            0: {"children": []}  # Root menu item is always first item in the menu
        }

        for item in self.menu_items:
            structure[item.id] = item.to_dbus()
            structure[0]["children"].append(item.id)

        return structure

    def _build_layout(
        self, parent_id, properties, menu: Optional[list[dict[int, dbus.Dictionary]]] = None
    ):
        """Build layout for GetLayout."""

        if menu is None:
            menu = self._build_menu_structure()

        if parent_id not in menu:
            return (
                dbus.Int32(parent_id),
                dbus.Dictionary({}, signature="sv"),
                dbus.Array([], signature="(ia{sv}av)")
            )

        item: dbus.Dictionary = menu[parent_id]

        children = dbus.Array([], signature="(ia{sv}av)")
        if "children" in item:
            for child_id in item["children"]:
                child_layout = self._build_layout(child_id, properties, menu)
                child_id_dbus = child_layout[0]
                child_props = child_layout[1] \
                    if child_layout[1] else dbus.Dictionary({}, signature="sv")
                child_children = child_layout[2] \
                    if child_layout[2] else dbus.Array([], signature="(ia{sv}av)")
                children.append(
                    dbus.Struct(
                        (child_id_dbus, child_props, child_children),
                        signature="(ia{sv}av)"
                    )
                )

        return (dbus.Int32(parent_id), item, children)

    @dbus.service.method(
        dbus_interface=DBUSMENU_INTERFACE,
        in_signature="iias",
        out_signature="u(ia{sv}av)"
    )
    def GetLayout(self, parent_id, _recursion_depth, property_names):  # pylint: disable=unused-argument, line-too-long, invalid-name # noqa: E501
        """Get menu layout"""
        layout = self._build_layout(parent_id, property_names)
        return (dbus.UInt32(self.revision), layout)

    @dbus.service.method(
        dbus_interface=DBUSMENU_INTERFACE,
        in_signature="aias",
        out_signature="a(ia{sv})"
    )
    def GetGroupProperties(self, ids, _property_names):  # pylint: disable=unused-argument, line-too-long, invalid-name # noqa: E501
        """Get properties for multiple items"""
        menu = self._build_menu_structure()
        result = dbus.Array([], signature="(ia{sv})")

        for item_id in ids:
            if item_id in menu:
                item: dbus.Dictionary = menu[item_id]

                result.append((dbus.Int32(item_id), item))

        return result

    @dbus.service.method(
        dbus_interface=DBUSMENU_INTERFACE,
        in_signature="isvu",
        out_signature=""
    )
    def Event(self, item_id, event_type, _data, _timestamp):  # pylint: disable=unused-argument, line-too-long, invalid-name # noqa: E501
        """Handle menu item clicks"""
        if event_type != "clicked":
            return

        item = [
            item for item in self.menu_items
            if item.id == item_id
            and item.type != MenuType.SEPARATOR.value
            and item.callback
        ]

        if item:
            GLib.idle_add(item.pop().callback)

    @dbus.service.signal(dbus_interface=DBUSMENU_INTERFACE, signature="ui")
    def LayoutUpdated(self, _revision, _parent):  # pylint: disable=unused-argument, invalid-name # noqa: E501
        """Update menu structure"""

    def update_menu(self):
        """Update menu structure"""
        self.revision += 1
        self.LayoutUpdated(self.revision, 0)


class _StatusNotifierItem(dbus.service.Object):
    """Internal StatusNotifierItem implementation"""

    def __init__(self, tray_icon, bus, object_path):
        self.tray = tray_icon
        self.bus = bus

        # Generate unique bus name
        self.bus_name_str = f"org.kde.StatusNotifierItem-{tray_icon.app_id}-{os.getpid()}"
        self.bus_name = dbus.service.BusName(self.bus_name_str, bus)

        super().__init__(self.bus_name, object_path)

        # Create DBusMenu
        self.menu = _DBusMenuService(bus, tray_icon.menu_items)

        # Register with watcher
        self._register_to_watcher()

    def _register_to_watcher(self):
        """Register with StatusNotifierWatcher"""
        try:
            watcher = self.bus.get_object(SNW_BUS_NAME, SNW_OBJECT_PATH)
            watcher.RegisterStatusNotifierItem(
                self.bus_name_str,
                dbus_interface=SNW_INTERFACE
            )
        except dbus.exceptions.DBusException:
            # Silent fail, will still work on some DEs,
            # see https://specifications.freedesktop.org/status-notifier-item-spec/status-notifier-item-spec-latest.html#registration  # pylint: disable=line-too-long # noqa: E501
            logger.warning(
                "Failed to register to StatusNotifierWatcher",
                category="TRAY_ICON",
                event="STATUS_NOTIFIER_WATCHER_REGISTRATION_FAILED",
            )

    @dbus.service.method(
        dbus_interface="org.freedesktop.DBus.Properties",
        in_signature="ss", out_signature="v"
    )
    def Get(self, interface: str, prop: str):  # pylint: disable=invalid-name
        """Get property"""
        if interface != SNI_INTERFACE:
            # Return an empty DBus string variant instead of None so the
            # dbus library can encode a valid value (None cannot be encoded).
            return dbus.String("")

        available_options = {
            StatusNotifierItemProperty.STATUS.value: dbus.String(self.tray.status),
            StatusNotifierItemProperty.CATEGORY.value: dbus.String("ApplicationStatus"),
            StatusNotifierItemProperty.ID.value: dbus.String(self.tray.app_id),
            StatusNotifierItemProperty.TITLE.value: dbus.String(self.tray.title),
            StatusNotifierItemProperty.ICON_NAME.value: dbus.String(self.tray.icon_name),
            StatusNotifierItemProperty.MENU.value: dbus.ObjectPath(DBUSMENU_PATH),
            StatusNotifierItemProperty.ITEM_IS_MENU.value: dbus.Boolean(True),
            StatusNotifierItemProperty.ICON_ACCESSIBLE_DESCRIPTION.value: dbus.String(
                self.tray.icon_desc
            ),
        }

        # Unknown property: return empty DBus string variant to avoid
        # "Don't know which D-Bus type to use to encode type NoneType" errors.
        return available_options.get(prop, "")

    @dbus.service.method(
        dbus_interface="org.freedesktop.DBus.Properties",
        in_signature="s",
        out_signature="a{sv}"
    )
    def GetAll(self, interface: str) -> dbus.Dictionary:  # pylint: disable=invalid-name
        """Get all properties"""
        if interface != SNI_INTERFACE:
            return {}

        return {
            StatusNotifierItemProperty.STATUS.value: dbus.String(self.tray.status),
            StatusNotifierItemProperty.CATEGORY.value: dbus.String("ApplicationStatus"),
            StatusNotifierItemProperty.ID.value: dbus.String(self.tray.app_id),
            StatusNotifierItemProperty.TITLE.value: dbus.String(self.tray.title),
            StatusNotifierItemProperty.ICON_NAME.value: dbus.String(self.tray.icon_name),
            StatusNotifierItemProperty.MENU.value: dbus.ObjectPath(DBUSMENU_PATH),
            StatusNotifierItemProperty.ITEM_IS_MENU.value: dbus.Boolean(True),
            StatusNotifierItemProperty.ICON_ACCESSIBLE_DESCRIPTION.value: dbus.String(
                self.tray.icon_desc
            )
        }

    @dbus.service.method(dbus_interface=SNI_INTERFACE, in_signature="ii", out_signature="")
    def Activate(self, _x_pos, _y_pos):  # pylint: disable=unused-argument, invalid-name, line-too-long # noqa: E501
        """Activate on click"""
        if self.tray.on_left_click:
            GLib.idle_add(self.tray.on_left_click)

    @dbus.service.signal(dbus_interface=SNI_INTERFACE, signature="")
    def NewIcon(self):  # pylint: disable=invalid-name
        """New icon"""

    def change_icon(self, icon_name, icon_desc):
        """Change icon"""
        self.tray.icon_name = icon_name
        self.tray.icon_desc = icon_desc
        self.NewIcon()


class TrayIcon:  # pylint: disable=too-many-instance-attributes
    """
    Easy-to-use tray icon with context menu for GTK4 apps

    Example:
        tray = TrayIcon(
            app_id="com.example.myapp",
            title="My Application",
            icon_name="application-x-executable"
        )

        tray.set_left_click(lambda: print("Clicked!"))
        tray.add_menu_item("Show", callback=show_window)
        tray.add_menu_separator()
        tray.add_menu_item("Quit", callback=quit_app)

        tray.setup()  # Call before app.run()
    """

    def __init__(
        self,
        app_id=APPLICATION_ID,
        title=TRAY_TITLE,
        icon_name=TRAY_ICON_NAME,
        icon_desc=TRAY_ICON_DESCRIPTION
    ):
        """
        Initialize tray icon

        Args:
            app_id: Application ID (e.g., "com.example.myapp")
            title: Tray icon tooltip/title
            icon_name: Icon name from theme (e.g., "application-x-executable")
        """
        self.app_id = app_id
        self.title = title
        self.icon_name = icon_name
        self.icon_desc = icon_desc
        self.status = "Active"

        self.menu_items = []
        self.on_left_click = None
        self.on_middle_click = None

        self._sni = None
        self._bus = None

    def set_left_click(self, callback):
        """
        Set callback for left-click on tray icon

        Args:
            callback: Function to call (no arguments)
        """
        self.on_left_click = callback

    def add_menu_item(self, label, callback, enabled=True, visible=True):
        """
        Add a menu item

        Args:
            label: Text to display
            callback: Function to call when clicked (no arguments)
            enabled: Whether item is clickable
            visible: Whether item is shown
        """
        self.menu_items.append(
            MenuObject(
                type=MenuType.ITEM,
                id=self._generate_menu_id(),
                label=label,
                callback=callback,
                enabled=enabled,
                visible=visible
            )
        )

    def add_menu_separator(self):
        """Add a separator line to the menu"""
        self.menu_items.append(
            MenuObject(
                type=MenuType.SEPARATOR,
                id=self._generate_menu_id(),
                label=""
            )
        )

    def _generate_menu_id(self):
        """Generate an used to identify the item in the menu."""
        return len(self.menu_items) + 1 if len(self.menu_items) > 0 else 1

    def setup(self):
        """
        Setup the tray icon

        IMPORTANT: Call this BEFORE app.run() in your GTK4 application!

        Example:
            app = MyGtkApp()
            tray = TrayIcon(...)
            tray.setup()
            app.run()
        """
        # Initialize D-Bus
        self._bus = dbus.SessionBus()

        # Create StatusNotifierItem which register itself to the StatusNotifierWatcher
        self._sni = _StatusNotifierItem(self, self._bus, SNI_PATH)

    def change_icon(self, icon_name, icon_desc):
        """
        Change the tray icon

        Args:
            icon_name: New icon name from theme
        """
        if self._sni:
            self._sni.change_icon(icon_name, icon_desc)

    def update_menu(self):
        """
        Update menu after adding/removing items dynamically

        Call this after modifying menu_items to refresh the menu
        """
        if self._sni:
            self._sni.menu.update_menu()



"""
Notification bar used to show error messages.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from enum import Enum
from typing import Optional

from gi.repository import GLib

from proton.vpn.app.gtk import Gtk


class NotificationType(Enum):
    """Types of notification messages"""
    SUCCESS = "success"
    ERROR = "error"
    INFO = "info"


class NotificationBar(Gtk.Revealer):
    """
    Notification bar widget used to show non-blocking error messages to the
    user for a brief period of time before they automatically disappear.
    """
    HIDE_NOTIFICATION_AFTER_MS = 10000

    def __init__(self):
        super().__init__()
        self._clear_error_message_src_id = None
        self._notification_label = Gtk.Label()
        self._notification_label.set_wrap(True)
        # set_max_width_chars is required for set_line_wrap to have effect.
        self._notification_label.set_max_width_chars(1)
        self._notification_label.set_justify(Gtk.Justification.CENTER)
        self.set_child(self._notification_label)

    @property
    def current_message(self):
        """Returns the notification message being shown."""
        return self._notification_label.get_label()

    def show_info_message(self, message: str, hide_after_ms: Optional[int] = None):
        """
        Shows the specified info message to the user for a limited amount of time.
        :param message: message to be shown.
        :param hide_after_ms: number of ms after which the error message will be hidden.
        The default value is NotificationBar.HIDE_NOTIFICATION_AFTER_MS.
        """
        self._show_notification(message, NotificationType.INFO, hide_after_ms)

    def show_error_message(self, message: str, hide_after_ms: Optional[int] = None):
        """
        Shows the specified error message to the user for a limited amount of time.
        :param message: error message to be shown.
        :param hide_after_ms: number of ms after which the error message will be hidden.
        The default value is NotificationBar.HIDE_NOTIFICATION_AFTER_MS.
        """
        self._show_notification(message, NotificationType.ERROR, hide_after_ms)

    def show_success_message(self, message: str, hide_after_ms: Optional[int] = None):
        """
        Shows the specified success message to the user for a limited amount of time.
        :param message: success message to be shown.
        :param hide_after_ms: number of ms after which the error message will be hidden.
        The default value is NotificationBar.HIDE_NOTIFICATION_AFTER_MS.
        """
        self._show_notification(message, NotificationType.SUCCESS, hide_after_ms)

    def clear(self):
        """Hides and clears the revealer."""
        self._clear_error_message()

    def _clear_error_message(self):
        self.set_reveal_child(False)
        self._notification_label.set_label("")
        self._notification_label.remove_css_class(NotificationType.SUCCESS.value)
        self._notification_label.remove_css_class(NotificationType.INFO.value)
        self.remove_css_class(NotificationType.ERROR.value)
        if self._clear_error_message_src_id:
            self._clear_error_message_src_id = None

        # We need to return False so that this function is only called once:
        # https://lazka.github.io/pgi-docs/#GLib-2.0/functions.html#GLib.timeout_add
        return False

    def _show_notification(
        self, message: str, notification_type: NotificationType,
        hide_after_ms: Optional[int] = None
    ):
        hide_after_ms = hide_after_ms or NotificationBar.HIDE_NOTIFICATION_AFTER_MS

        if self._clear_error_message_src_id:
            # Remove the source that will clear the previous error message,
            # otherwise, it might remove the error message soon after it's added.
            GLib.source_remove(self._clear_error_message_src_id)
            self._clear_error_message_src_id = None

        self._notification_label.set_label(message)

        if notification_type == NotificationType.ERROR:
            self._notification_label.remove_css_class(NotificationType.SUCCESS.value)
            self._notification_label.remove_css_class(NotificationType.INFO.value)
        elif notification_type == NotificationType.SUCCESS:
            self._notification_label.remove_css_class(NotificationType.INFO.value)
            self._notification_label.remove_css_class(NotificationType.ERROR.value)
        else:
            self._notification_label.remove_css_class(NotificationType.ERROR.value)
            self._notification_label.remove_css_class(NotificationType.SUCCESS.value)

        self._notification_label.add_css_class(notification_type.value)
        self.set_reveal_child(True)

        self._clear_error_message_src_id = GLib.timeout_add(
            hide_after_ms, self._clear_error_message,
        )



"""
This module defines the login widget, used to authenticate the user.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import TYPE_CHECKING, Optional

from gi.repository import GObject

from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk import Gtk
from proton.vpn import logging
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.login.login_stack import LoginStack
from proton.vpn.app.gtk.widgets.login.disable_killswitch import DisableKillSwitchWidget
from proton.vpn.app.gtk.widgets.main.notifications import Notifications
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget
from proton.vpn.connection.enum import KillSwitchSetting as KillSwitchSettingEnum

if TYPE_CHECKING:
    from proton.vpn.app.gtk.app import MainWindow


logger = logging.getLogger(__name__)


#  pylint: disable=too-many-arguments
class LoginWidget(Gtk.Box):
    """Container widget that holds both
    login UI and also a revealer to display
    in the case the user is logged out and permanent
    kill switch is enabled.
    """
    def __init__(
        self,
        controller: Controller,
        notifications: Notifications,
        overlay_widget: OverlayWidget,
        main_window: "MainWindow",
        login_stack: Optional["LoginStack"] = None,
        disable_killswitch_widget: Optional[DisableKillSwitchWidget] = None
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        self.set_name("login-widget")

        self._controller = controller

        self.login_stack = login_stack or LoginStack(
            self._controller, notifications, overlay_widget
        )
        safe_signal_connect(self.login_stack, "user-logged-in", self._on_user_logged_in)

        self.disable_killswitch = disable_killswitch_widget or DisableKillSwitchWidget(
            main_window
        )
        safe_signal_connect(
            self.disable_killswitch,
            "disable-killswitch",
            self._on_disable_killswitch
        )

        self.append(self.login_stack)
        self.append(self.disable_killswitch)

    @GObject.Signal
    def user_logged_in(self):
        """Signal emitted after a successful login."""

    def _on_user_logged_in(self, _: "LoginStack"):
        self.emit("user-logged-in")

    def reset(self):
        """Proxy method to reset the widget to its initial state."""
        is_ks_permanent = self._controller.get_settings()\
            .killswitch == KillSwitchSettingEnum.PERMANENT
        self.disable_killswitch.set_reveal_child(is_ks_permanent)
        self.login_stack.login_form.set_property("sensitive", not is_ks_permanent)
        self.login_stack.reset()

    def _on_disable_killswitch(self, _):
        settings = self._controller.get_settings()
        settings.killswitch = KillSwitchSettingEnum.OFF
        self._controller.save_settings(settings)

        self.disable_killswitch.set_reveal_child(False)
        self.login_stack.login_form.set_property("sensitive", True)



"""
Server search results module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Callable, Iterable, Optional, Set, Tuple

from gi.repository import GObject

from proton.vpn.session.servers.logicals import (
    sort_servers_alphabetically_by_country_and_server_name
)

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn import logging

logger = logging.getLogger(__name__)

# We're hittig a bug in the GTK treeview where x11 crashes with BadAlloc
# if we try to display > 1500 rows in certain configurations.
# This avoids the issue by limiting the number of results displayed.
# This is ultimately a better user experience as the whole point of
# the search is to find a specific item.
MAX_SEARCH_RESULTS_PER_SECTION = 100

# Column indexes in the tree model for the FilteredList.
COLUMN_NAME = 0  # The name of the item.
COLUMN_LOAD = 1  # The server load
COLUMN_LOAD_COLOR = 2  # Sets the column color.
COLUMN_SENSITIVE = 3  # Whether the item is sensitive to selection.
LOAD_COLOR = "Grey"  # We want to show the load % in grey.


class FilteredList(Gtk.TreeView):
    """
    Displays a list of countries, locations and servers in a tree view.
    """
    def __init__(
        self,
        countries: Callable[[Optional[str]], Iterable[Tuple[Optional[str], Optional[int]]]],
        locations: Callable[[Optional[str]], Iterable[Tuple[Optional[str], Optional[int]]]],
        servers: Callable[[Optional[str]], Iterable[Tuple[Optional[str], Optional[int]]]]
    ):
        super().__init__()
        self._countries = countries
        self._locations = locations
        self._servers = servers
        self._model = Gtk.TreeStore(str, str, str, bool)
        self.set_model(Gtk.TreeModelSort(model=self._model))

        self.set_show_expanders(False)
        self.set_activate_on_single_click(True)

        def select_function(
            _treeselection: Gtk.TreeSelection,
            model: Gtk.TreeModel,
            path: Gtk.TreePath,
            _current: bool
        ) -> bool:
            tree_iter = model.get_iter(path)
            if not tree_iter:
                return False

            sensitivity = model.get_value(tree_iter, COLUMN_SENSITIVE)
            return sensitivity

        self.get_selection().set_select_function(select_function)

        # server name
        name_column = Gtk.TreeViewColumn(
            "Name",
            cell_renderer=Gtk.CellRendererText(),
            text=COLUMN_NAME,
            sensitive=COLUMN_SENSITIVE
        )

        self.append_column(name_column)

        # server load
        load_column = Gtk.TreeViewColumn(
            "Load",
            cell_renderer=Gtk.CellRendererText(),
            text=COLUMN_LOAD,
            sensitive=COLUMN_SENSITIVE,
            foreground=COLUMN_LOAD_COLOR
        )

        self.append_column(load_column)

        self.set_headers_visible(False)

    def update(self, search_text: Optional[str] = None):
        """Rebuild the view using the search_text as a filter"""
        self._model.clear()

        sections = (
            ("Countries", self._countries, ""),
            ("Locations", self._locations, ""),
            ("Servers", self._servers, "Server load")
        )

        for section in sections:
            section_name, section_data, section_load_header = section
            data = list(section_data(search_text))
            if not data:
                continue

            row = [f"{section_name} ({len(data)})", section_load_header, LOAD_COLOR, False]
            root = self._model.append(None, row)
            for i, (name, load) in enumerate(data):
                load_string = "" if load is None else f"{load}%"
                self._model.append(
                    root,
                    [name, load_string, LOAD_COLOR, True]
                )
                if i == MAX_SEARCH_RESULTS_PER_SECTION:
                    self._model.append(
                        None,
                        ["...", "", LOAD_COLOR, False]
                    )
                    break

        self.expand_all()


class SearchResults(Gtk.ScrolledWindow):
    """Display a filtered view of countries and servers.
       Inside a scroll-able widget.
    """
    def __init__(self, controller):
        super().__init__()
        self.set_name("search-results")
        self.set_policy(
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            vscrollbar_policy=Gtk.PolicyType.AUTOMATIC
        )
        self.set_vexpand(False)
        self._container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.set_child(self._container)
        self.set_property("height-request", 200)

        self._revealer: Optional[Gtk.Revealer] = None

        def countries(search_text: Optional[str] = None) -> Set[Tuple[Optional[str], None]]:
            result: Set[Tuple[Optional[str], None]] = set()
            server_list = controller.server_list

            if not server_list:
                return result

            for server in controller.server_list:
                if not server.under_maintenance and \
                        self._search_input_exists(search_text, server, entry_country_name=True):
                    result.add((server.entry_country_name, None))

            return result

        def locations(
            search_text: Optional[str] = None
        ) -> Iterable[Tuple[Optional[str], Optional[int]]]:
            result = set({})
            server_list = controller.server_list

            if not server_list:
                yield (None, None)
                return

            for server in controller.server_list:
                if not server.under_maintenance and \
                        self._search_input_exists(search_text, server, location_name=True):
                    if server.location:
                        result.add(server.location)

            for location in sorted(result):
                yield (location, None)

        def servers(
            search_text: Optional[str] = None
        ) -> Iterable[Tuple[Optional[str], Optional[int]]]:
            def user_tier_allows_access_to_server(server_tier: int, user_tier: int) -> bool:
                return server_tier <= user_tier

            user_tier = controller.user_tier
            server_list = controller.server_list
            if not server_list:
                yield (None, None)
                return

            matches = [
                server for server in server_list
                if not server.under_maintenance
                and user_tier_allows_access_to_server(server.tier, user_tier)
                and self._search_input_exists(search_text, server, server_name=True)
            ]
            for server in sorted(
                matches,
                key=sort_servers_alphabetically_by_country_and_server_name
            ):
                yield (server.name, server.load)

        self._filtered_country_list = FilteredList(countries, locations, servers)
        safe_signal_connect(
            self._filtered_country_list,
            "row-activated", self._on_row_activated
        )

        self._container.append(self._filtered_country_list)

    def set_revealer(self, revealer: Optional[Gtk.Revealer]):
        """Sets revealer for search results"""
        self._revealer = revealer

    def _search_input_exists(  # pylint: disable=too-many-arguments
        self, search_text: Optional[str], server, entry_country_name: bool = False,
        location_name: bool = False, server_name: bool = False
    ) -> bool:
        if entry_country_name:
            return bool(search_text and (search_text in server.entry_country_name.lower()))

        if location_name:
            return bool(search_text and (search_text in server.location.lower()))

        if server_name:
            return bool(search_text and (search_text in server.name.lower()))

        return False

    @GObject.Signal(name="result-chosen", arg_types=(str,))
    def result_chosen(self, _row: str):
        """Broadcast that a result has been chosen in the search results."""

    def on_search_changed(self, search_widget: Gtk.SearchEntry):
        """Callback when search entry has changed."""
        search_text = search_widget.get_text().lower()

        self._filtered_country_list.update(search_text)
        if self._revealer:
            self._revealer.set_reveal_child(bool(search_text))

    def _on_row_activated(
        self,
        tree_view: FilteredList,
        path: Gtk.TreePath,
        _tree_view_column: Gtk.TreeViewColumn
    ):
        model = tree_view.get_model()
        tree_iter = model.get_iter(path)
        if tree_iter:
            selected_value = model.get_value(tree_iter, COLUMN_NAME)
            self._revealer.set_reveal_child(False)
            self.emit("result-chosen", selected_value)
